from typing import Literal

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from sklearn.ensemble import ExtraTreesClassifier as SKExtraTreesClassifier
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
This is an sklearn extra trees classifier model.

The extra trees classifier is a meta estimator that fits a number of randomized decision
trees (a.k.a. extra-trees) on various sub-samples of the dataset and uses averaging to
improve the predictive accuracy and control over-fitting.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
n_estimators : int, default=100
    The number of trees in the forest.
criterion : {"gini", "entropy", "log_loss"}, default="gini"
    The function to measure the quality of a split. Supported criteria are
    "gini" for the Gini impurity and "log_loss" and "entropy" both for the
    Shannon information gain.
    Note: This parameter is tree-specific.
max_depth : int, default=None
    The maximum depth of the tree. If None, then nodes are expanded until
    all leaves are pure or until all leaves contain less than
    min_samples_split samples.
min_samples_split : int or float, default=2
    The minimum number of samples required to split an internal node:
    - If int, then consider `min_samples_split` as the minimum number.
    - If float, then `min_samples_split` is a fraction and
      `ceil(min_samples_split * n_samples)` are the minimum
      number of samples for each split.
min_samples_leaf : int or float, default=1
    The minimum number of samples required to be at a leaf node.
    A split point at any depth will only be considered if it leaves at
    least ``min_samples_leaf`` training samples in each of the left and
    right branches.  This may have the effect of smoothing the model,
    especially in regression.
    - If int, then consider `min_samples_leaf` as the minimum number.
    - If float, then `min_samples_leaf` is a fraction and
      `ceil(min_samples_leaf * n_samples)` are the minimum
      number of samples for each node.
min_weight_fraction_leaf : float, default=0.0
    The minimum weighted fraction of the sum total of weights (of all
    the input samples) required to be at a leaf node. Samples have
    equal weight when sample_weight is not provided.
max_features : {"sqrt", "log2", None}, int or float, default="sqrt"
    The number of features to consider when looking for the best split:
    - If int, then consider `max_features` features at each split.
    - If float, then `max_features` is a fraction and
      `round(max_features * n_features)` features are considered at each
      split.
    - If "auto", then `max_features=sqrt(n_features)`.
    - If "sqrt", then `max_features=sqrt(n_features)`.
    - If "log2", then `max_features=log2(n_features)`.
    - If None, then `max_features=n_features`.
    Note: the search for a split does not stop until at least one
    valid partition of the node samples is found, even if it requires to
    effectively inspect more than ``max_features`` features.
max_leaf_nodes : int, default=None
    Grow trees with ``max_leaf_nodes`` in best-first fashion.
    Best nodes are defined as relative reduction in impurity.
    If None then unlimited number of leaf nodes.
min_impurity_decrease : float, default=0.0
    A node will be split if this split induces a decrease of the impurity
    greater than or equal to this value.
    The weighted impurity decrease equation is the following::
        N_t / N * (impurity - N_t_R / N_t * right_impurity
                            - N_t_L / N_t * left_impurity)
    where ``N`` is the total number of samples, ``N_t`` is the number of
    samples at the current node, ``N_t_L`` is the number of samples in the
    left child, and ``N_t_R`` is the number of samples in the right child.
    ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
    if ``sample_weight`` is passed.
bootstrap : bool, default=True
    Whether bootstrap samples are used when building trees. If False, the
    whole dataset is used to build each tree.
oob_score : bool, default=False
    Whether to use out-of-bag samples to estimate the generalization score.
    Only available if bootstrap=True.
n_jobs : int, default=None
    The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
    :meth:`decision_path` and :meth:`apply` are all parallelized over the
    trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
    context. ``-1`` means using all processors.
random_state : int or None, default=None
    Controls both the randomness of the bootstrapping of the samples used
    when building trees (if ``bootstrap=True``) and the sampling of the
    features to consider when looking for the best split at each node
    (if ``max_features < n_features``).
verbose : int, default=0
    Controls the verbosity when fitting and predicting.
warm_start : bool, default=False
    When set to ``True``, reuse the solution of the previous call to fit
    and add more estimators to the ensemble, otherwise, just fit a whole
    new forest.
class_weight : {"balanced", "balanced_subsample"}, dict or list of dicts, \
        default=None
    Weights associated with classes in the form ``{class_label: weight}``.
    If not given, all classes are supposed to have weight one. For
    multi-output problems, a list of dicts can be provided in the same
    order as the columns of y.
    Note that for multioutput (including multilabel) weights should be
    defined for each class of every column in its own dict. For example,
    for four-class multilabel classification weights should be
    [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
    [{1:1}, {2:5}, {3:1}, {4:1}].
    The "balanced" mode uses the values of y to automatically adjust
    weights inversely proportional to class frequencies in the input data
    as ``n_samples / (n_classes * np.bincount(y))``
    The "balanced_subsample" mode is the same as "balanced" except that
    weights are computed based on the bootstrap sample for every tree
    grown.
    For multi-output, the weights of each column of y will be multiplied.
    Note that these weights will be multiplied with sample_weight (passed
    through the fit method) if sample_weight is specified.
ccp_alpha : non-negative float, default=0.0
    Complexity parameter used for Minimal Cost-Complexity Pruning. The
    subtree with the largest cost complexity that is smaller than
    ``ccp_alpha`` will be chosen. By default, no pruning is performed.
max_samples : int or float, default=None
    If bootstrap is True, the number of samples to draw from X
    to train each base estimator.
    - If None (default), then draw `X.shape[0]` samples.
    - If int, then draw `max_samples` samples.
    - If float, then draw `max_samples * X.shape[0]` samples. Thus,
      `max_samples` should be in the interval `(0.0, 1.0]`.
"""

ClassWeights = Literal["balanced", "balanced_subsample"]
Criterion = Literal["gini", "entropy", "log_loss"]
MaxFeaturesCallable = Literal["sqrt", "log2"]


class Config:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


@dataclasses.dataclass(config=Config)
class ExtraTreesClassifierConfig(ModelConfig):
    n_estimators: int = 100
    criterion: Criterion = "gini"
    max_depth: int | None = None
    min_samples_split: int | float = 2
    min_samples_leaf: int | float = 1
    min_weight_fraction_leaf: float = 0.0
    max_features: float | MaxFeaturesCallable | None = "sqrt"
    max_leaf_nodes: int | None = None
    min_impurity_decrease: float = 0.0
    bootstrap: bool = False
    oob_score: bool = False
    n_jobs: int | None = None
    random_state: int | None = None
    verbose: int = 0
    warm_start: bool = False
    class_weight: ClassWeights | dict | list[dict] | None = None
    ccp_alpha: float = 0.0
    max_samples: int | float | None = None

    def __post_init_post_parse__(self) -> None:
        if self.ccp_alpha < 0:
            raise ValueError("Complexity parameter `ccp_alpha` cannot be negative.")


class ExtraTreesClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[ExtraTreesClassifierConfig],
):
    @property
    def _config_builder(self) -> type[ExtraTreesClassifierConfig]:
        return ExtraTreesClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKExtraTreesClassifier:
        config = self.model_config
        return SKExtraTreesClassifier(
            n_estimators=config.n_estimators,
            criterion=config.criterion,
            max_depth=config.max_depth,
            min_samples_split=config.min_samples_split,
            min_samples_leaf=config.min_samples_leaf,
            min_weight_fraction_leaf=config.min_weight_fraction_leaf,
            max_features=config.max_features,
            max_leaf_nodes=config.max_leaf_nodes,
            min_impurity_decrease=config.min_impurity_decrease,
            bootstrap=config.bootstrap,
            oob_score=config.oob_score,
            n_jobs=config.n_jobs,
            random_state=config.random_state,
            verbose=config.verbose,
            warm_start=config.warm_start,
            class_weight=config.class_weight,
            ccp_alpha=config.ccp_alpha,
            max_samples=config.max_samples,
        )



from molflux.splits.catalogue import (
    fill_catalogue,
    list_splitting_strategies,
    register_splitting_strategy,
)
from molflux.splits.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_splitting_strategy,
)
from molflux.splits.presets import (
    k_fold_split,
    train_test_split,
    train_validation_test_split,
)
from molflux.splits.strategy import SplittingStrategy

# Register all plugins at package import time (to fill catalogue)
fill_catalogue()



"""Computes a diversity ROC curve over SMILES string."""

import logging
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@InProceedings{exscientia:diversity_roc,
title = {Diversity ROC},
authors={Exscientia},
year={2022}
}
"""

_DESCRIPTION = """\
Computes a diversity ROC curve over SMILES strings, i.e. "how many SMILES in the first k results are diverse"?
"""


# TODO: Add description of the arguments of the metric here
_KWARGS_DESCRIPTION = """
Calculates what proportion of SMILES-encoded predictions are diverse within the top k tries.

Args:
    predictions: list of lists of predictions to score. Each prediction should be a SMILES string.
    references:  list of references. Ignored by this metric, but should be of the same length as 'predictions'.
    average: If `True`, an molflux-wise average across results is performed.

Returns:
    diversity_roc: list whose k component is the answer to "how many diverse strings are in top-k predictions".
        If 'average=True', "on average, how many diverse strings are in top-k predictions".

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("diversity_roc")
    >>> predictions = [["CC", "C", "CC", "C", "CCC"]]
    >>> references = [None]
    >>> metric.compute(predictions=predictions, references=references)
    {'diversity_roc': [[1.0, 1.0, 0.66..., 0.5, 0.6]]}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class DiversityRoc(HFMetric):
    """Computes the diversity ROC curve."""

    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            # This is the description that will appear on the metrics page.
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            # This defines the format of each prediction
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("string")),
                    "references": datasets.Value("string"),
                },
            ),
        )

    def _pre_process_inputs(
        self,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
    ) -> tuple[ArrayLike | None, ...]:
        """Converts None inputs to strings to allow using None as invalid token.

        Otherwise, None molfluxs would raise as they do not count as string features.
        """
        if predictions is not None:
            predictions = [
                [p if p is not None else "None" for p in ps] for ps in predictions
            ]

        if references is not None:
            references = [r if r is not None else "None" for r in references]

        return predictions, references

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        average: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        """Returns the diversity ROC curve"""

        max_k = len(predictions[0])

        # vector of 0./1. representing whether a prediction is repeated
        def mark_new_predictions(preds: list[str]) -> np.ndarray:
            idxs = sorted([preds.index(p) for p in set(preds)])
            new_predictions = np.zeros((max_k,))
            new_predictions[idxs] = 1
            return new_predictions

        new_predictions = [mark_new_predictions(preds) for preds in predictions]

        def running_mean(x: np.ndarray) -> np.ndarray:
            cumsum = np.cumsum(x)
            norm = np.arange(1, np.shape(x)[0] + 1)
            mean: np.ndarray = cumsum / norm
            return mean

        running_diversities = np.array([running_mean(p) for p in new_predictions])

        if average:
            running_diversities = running_diversities.mean(axis=0, keepdims=True)

        diversity_roc: list[float] = running_diversities.tolist()

        return {self.tag: diversity_roc}



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.rdkit.fingerprints.mhfp_unfolded import (
    MHFPUnfolded,
)

representation_name = "mhfp_unfolded"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, MHFPUnfolded)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_compute(fixture_representation):
    """That default scoring gives expected results.

    This test is based on the original mhfp package test suite.
    """
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples, n_permutations=32, seed=42)
    expected_result = [
        4020624,
        9980344,
        63689098,
        22015504,
        26474917,
        18400837,
        47464384,
        217139116,
        8836027,
        41227994,
        3838604,
        13228201,
        46316417,
        3763579,
        38514427,
        37657111,
        55411077,
        223018370,
        43121840,
        13101872,
        39382703,
        31834049,
        11797863,
        158268734,
        12619637,
        17796484,
        5277139,
        41610747,
        106667704,
        17532210,
        52062080,
        147129299,
    ]
    assert "mhfp_unfolded" in result
    assert result["mhfp_unfolded"] == [expected_result]



import pytest

from molflux.features.errors import DuplicateKeyError
from molflux.features.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_representation,
)
from molflux.features.representation import Representation, Representations

representative_representation_name = "character_count"


def test_returns_representation():
    """That loading a representation returns an object of type Representation."""
    name = representative_representation_name
    representation = load_representation(name=name)
    assert isinstance(representation, Representation)


def test_raises_not_implemented_for_unknown_representation():
    """That a NotImplementedError is raised if attempting to load an unknown
    representation."""
    name = "unknown"
    with pytest.raises(NotImplementedError):
        load_representation(name=name)


def test_raised_error_for_unknown_representation_provides_close_matches():
    """That the error raised when attempting to load an unknown representation
    shows possible close matches to the user (if any)."""
    name = "character_cont"
    # This should suggest e.g. ["character_count"]
    with pytest.raises(
        NotImplementedError,
        match="You might be looking for one of these",
    ):
        load_representation(name=name)


def test_forwards_init_kwargs_to_builder():
    """That keyword arguments get forwarded to the representation initialiser."""
    name = representative_representation_name
    representation = load_representation(name=name, tag="pytest-tag")
    assert representation.tag == "pytest-tag"


def test_from_dict_returns_representation():
    """That loading from a dict returns a Representation."""
    name = representative_representation_name
    config = {
        "name": name,
        "config": {},
        "presets": {},
    }
    representation = load_from_dict(config)
    assert isinstance(representation, Representation)


def test_from_minimal_dict():
    """That can provide a config with only required fields."""
    name = representative_representation_name
    config = {
        "name": name,
    }
    assert load_from_dict(config)


def test_from_dict_forwards_config_to_builder():
    """That config keyword arguments get forwarded to the initialiser."""
    name = representative_representation_name
    config = {
        "name": name,
        "config": {
            "tag": "pytest-tag",
        },
    }
    representation = load_from_dict(config)
    assert representation.tag == "pytest-tag"


def test_from_dict_forwards_presets_to_representation_state():
    """That presets keyword arguments get stored in the representation."""
    name = representative_representation_name
    config = {
        "name": name,
        "presets": {
            "without_hs": True,
        },
    }
    representation = load_from_dict(config)
    assert representation.state
    assert representation.state.get("without_hs") is True


def test_dict_missing_required_fields_raises():
    """That cannot load a representation with a config missing required fields."""
    config = {"unknown_key": "value"}
    with pytest.warns(UserWarning), pytest.raises(SyntaxError):
        load_from_dict(config)


def test_representations_with_same_tag_in_collection_raises():
    """That adding several representations with the same tag in a Representations collection raises."""
    config = {
        "name": representative_representation_name,
    }
    duplicate_configs = [config, config]
    with pytest.raises(DuplicateKeyError):
        load_from_dicts(duplicate_configs)


def test_load_from_dicts_returns_correct_number_of_representations():
    """That loading from dicts returns a Representations collection of the
    expected size."""
    name = representative_representation_name
    config_one = {
        "name": name,
        "config": {
            "tag": "one",
        },
        "presets": {},
    }
    config_two = {
        "name": name,
        "config": {
            "tag": "two",
        },
        "presets": {},
    }
    configs = [config_one, config_two]
    representations = load_from_dicts(configs)
    assert len(representations) == 2


def test_from_yaml_returns_representations(fixture_path_to_assets):
    path = fixture_path_to_assets / "config.yml"
    representations = load_from_yaml(path=path)
    assert len(representations) == 2
    assert "character_count" in representations
    assert "custom_descriptor" in representations
    assert isinstance(representations, Representations)
    assert representations["custom_descriptor"].state.get("without_hs") is True



import os
from copy import copy
from dataclasses import asdict
from typing import Any

import numpy as np
from pydantic.v1 import dataclasses

from datasets import Dataset
from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.load import load_from_dict
from molflux.modelzoo.model import ClassificationMixin, ModelBase, ModelConfig
from molflux.modelzoo.models.ensemble._combo.utils import (
    check_parameter,
    get_split_indices,
    list_diff,
)
from molflux.modelzoo.protocols import SupportsClassification
from molflux.modelzoo.typing import Classes, PredictionResult
from molflux.modelzoo.utils import format_wrapped_model_tag, pick_features

try:
    from molflux.modelzoo.models.sklearn.logistic_regressor import LogisticRegressor
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
Implementation of an ensemble classification model.

An ensemble of models is a collection of models trained separately on the same data that
when combined together may offer better performance than any of the individual models.
Ensemble models often perform well on competition leaderboards, even when ensembling
lots of models that individually may perform poorly.

In this implementation of meta-ensembling (also known as stacking), each of the
base classification models makes a prediction and a meta-classifier is trained
on the predictions of the individual models, optionally including the original
data as additional features.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
base_estimators : list
    A list of modelzoo config dicts to define the models used for training
    the base estimators. An error is raised if the base estimators are not
    provided.
meta_estimator : dict, modelzoo model config, optional (default=logistic regression)
    The config for a modelzoo meta classifier to make the final prediction.
n_folds : int, optional (default=2)
    The number of splits of the training sample.
keep_original : bool, optional (default=False)
    If True, keep the original features for training and predicting.
use_proba: bool, optional (default=False)
    If True, use the probability prediction as the new features.
shuffle_data : bool, optional (default=False)
    If True, shuffle the input data.
random_state : int, RandomState or None, optional (default=None)
    If int, random_state is the seed used by the random
    number generator; If RandomState instance, random_state is the random
    number generator; If None, the random number generator is the
    RandomState instance used by `np.random`.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class EnsembleClassifierConfig(ModelConfig):
    base_estimators: list[dict[str, Any]] | None = None
    meta_estimator: dict[str, Any] | None = (
        None  # config for a a modelzoo classification model
    )
    n_folds: int = 2
    keep_original: bool = False
    use_proba: bool = False
    shuffle_data: bool = False
    random_state: int | None = None


class EnsembleClassifier(ClassificationMixin, ModelBase[EnsembleClassifierConfig]):
    """
    Meta ensembling, also known as stacking
    See https://datasciblog.github.io/2016/12/27/a-kagglers-guide-to-model-stacking-in-practice/ for more information.
    See also the `combo` library eg. https://github.com/yzhao062/combo/blob/master/combo/models/classifier_stacking.py
    """

    def __init__(self, tag: str | None = None, **config_kwargs: Any) -> None:
        super().__init__(tag=tag, **config_kwargs)

        config = self.config
        if len(config["base_estimators"]) < 2:
            raise ValueError("At least 2 estimators are required")
        self.base_estimators = [
            load_from_dict(cfg) for cfg in config["base_estimators"]
        ]
        self.n_base_estimators_ = len(self.base_estimators)
        if tag is None:
            self.info.tag = format_wrapped_model_tag(
                self.info.tag,
                [estimator.tag for estimator in self.base_estimators],
            )

        # validate input parameters
        if not isinstance(config["n_folds"], int):
            raise ValueError("n_folds must be an integer variable")
        check_parameter(
            config["n_folds"],
            low=2,
            include_left=True,
            param_name="n_folds",
        )
        self.n_folds = config["n_folds"]

        if config["meta_estimator"] is not None:
            self.meta_estimator = load_from_dict(config["meta_estimator"])
        elif config["keep_original"]:
            meta_estimator_colnames = [
                clf.tag for clf in self.base_estimators
            ] + self.x_features
            self.meta_estimator = LogisticRegressor(
                x_features=meta_estimator_colnames,
                y_features=self.y_features,
            )
        elif not config["keep_original"]:
            meta_estimator_colnames = [clf.tag for clf in self.base_estimators]
            self.meta_estimator = LogisticRegressor(
                x_features=meta_estimator_colnames,
                y_features=self.y_features,
            )
        if not isinstance(self.meta_estimator, SupportsClassification):
            raise ValueError("Meta estimator should support classification")

        # set flags
        self.keep_original = config["keep_original"]
        self.shuffle_data = config["shuffle_data"]
        self.random_state = config["random_state"]
        self.use_proba = config["use_proba"]

    def _config(self) -> EnsembleClassifierConfig:
        return EnsembleClassifierConfig()

    @property
    def config(self) -> dict[str, Any]:
        return asdict(self.model_config)

    @property
    def _config_builder(self) -> type[EnsembleClassifierConfig]:
        return EnsembleClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _train(self, train_data: Dataset, **kwargs: Any) -> Any:
        """
        Fit ensemble classifier. Iterate over folds of cross validation set to fit individual ensemble models.
        Fit meta-estimator based on these. Then refit individual models on full data.
        """
        n_samples = train_data.shape[0]

        # initialize matrix for storing newly generated features
        new_features = np.zeros([n_samples, self.n_base_estimators_])

        # build CV datasets
        index_lists = get_split_indices(
            dataset=train_data,
            n_folds=self.n_folds,
        )
        if self.shuffle_data:
            train_data = train_data.shuffle(seed=self.random_state)

        # iterate over all base estimators
        for i, clf in enumerate(self.base_estimators):
            if not isinstance(clf, SupportsClassification):
                raise ValueError("Base estimators should support classification")
            # iterate over all folds
            for j in range(self.n_folds):
                # build train and test index
                full_idx = list(range(n_samples))
                test_idx = index_lists[j]
                train_idx = list_diff(full_idx, test_idx)
                train_fold = train_data.select(train_idx)
                test_fold = train_data.select(test_idx)

                # train the classifier
                clf.train(train_data=train_fold)

                # generate the new features on the pseudo test set
                clf_tag_name = f"{clf.tag}::{clf.y_features[0]}"
                if self.use_proba:
                    new_features[test_idx, i] = clf.predict_proba(test_fold)[
                        clf_tag_name + "::probabilities"
                    ]
                else:
                    new_features[test_idx, i] = clf.predict(test_fold)[clf_tag_name]

        # build the new dataset for training
        train_comb = copy(train_data)
        meta_estimator_colnames = [clf.tag for clf in self.base_estimators]
        for i_col, colname in enumerate(meta_estimator_colnames):
            train_comb = train_comb.add_column(
                colname,
                new_features[:, i_col],
            ).flatten_indices()
        if self.keep_original:
            meta_estimator_colnames += self.x_features
        assert self.meta_estimator.x_features == meta_estimator_colnames
        self.meta_estimator.train(train_data=train_comb)

        # train all base classifiers on the full train dataset
        # iterate over all base estimators
        for clf in self.base_estimators:
            clf.train(train_data=train_data)

    def _process_data(self, dataset: Dataset) -> Any:
        """Internal class for `predict`
        Parameters
        ----------
        dataset: Dataset (n_samples, n_features)
            The input samples.
        Returns
        -------
        data_new_comb : Dataset
            The processed dataset
        """
        n_samples = dataset.shape[0]

        # initialize matrix for storing newly generated features
        new_features = np.zeros([n_samples, self.n_base_estimators_])

        # build the new features for unknown samples
        # iterate over all base classifiers
        for i, clf in enumerate(self.base_estimators):
            clf_tag_name = f"{clf.tag}::{clf.y_features[0]}"
            if self.use_proba:
                new_features[:, i] = clf.predict_proba(dataset)[  # type: ignore[attr-defined]
                    clf_tag_name + "::probabilities"
                ]
            else:
                new_features[:, i] = clf.predict(dataset)[clf_tag_name]

        # build the new dataset for unknown samples
        new_colnames = [clf.tag for clf in self.base_estimators]
        data_new_comb = copy(dataset)
        for i_col, colname in enumerate(new_colnames):
            data_new_comb = data_new_comb.add_column(
                colname,
                new_features[:, i_col],
            ).flatten_indices()

        return data_new_comb

    def _predict(self, data: Dataset, **kwargs: Any) -> PredictionResult:
        """Predict class labels for the provided data.
        Parameters
        ----------
        data: Dataset
        Returns
        -------
        labels : numpy array of shape (n_samples,)
            Predictions for each data sample.
        """
        display_names = self._predict_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X_new_comb = self._process_data(x_data)
        original_prediction_results = self.meta_estimator.predict(X_new_comb)
        return dict(
            zip(
                display_names,
                original_prediction_results.values(),
                strict=False,
            ),
        )

    @property
    def classes(self) -> Classes:
        if not hasattr(self.model, "classes_"):
            return {}

        sklearn_classes = self.model.classes_
        # this could be of several shapes / types
        # singletask: (n_classes,)
        # multitask: (n_tasks, n_classes)

        is_multitask = np.ndim(sklearn_classes) == 2
        classes = sklearn_classes if is_multitask else [sklearn_classes]

        return {
            task: classes.tolist()
            for task, classes in zip(self.y_features, classes, strict=False)
        }

    def _predict_proba(self, data: Dataset, **kwargs: Any) -> PredictionResult:
        """Return probability estimates for the test data X.
        Parameters
        ----------
        data: Dataset
        Returns
        -------
        p : numpy array of shape (n_samples,)
            The class probabilities of the input samples.
            Classes are ordered by lexicographic order.
        """
        display_names = self._predict_proba_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        x_data = pick_features(data, self.x_features)
        X_new_comb = self._process_data(x_data)
        original_prediction_results = self.meta_estimator.predict_proba(X_new_comb)  # type: ignore[attr-defined]
        return dict(
            zip(
                display_names,
                original_prediction_results.values(),
                strict=False,
            ),
        )

    def as_dir(self, directory: str) -> None:
        """
        Serialises a pre-trained ensemble model in a directory.

        As ensembles have a meta estimator and a list of base estimators, each one is
        saved to a subdirectory, in the format:
        `directory/
            meta_estimator/
                ...
            base_estimators/
                0/
                    ...
                1/
                    ...
                ...
        """

        if self.meta_estimator is None:
            raise NotTrainedError

        if self.base_estimators is None:
            raise NotTrainedError

        meta_estimator_path = os.path.join(directory, "meta_estimator")
        base_estimators_base_path = os.path.join(directory, "base_estimators")

        save_to_store(meta_estimator_path, self.meta_estimator)

        for model_index in range(len(self.base_estimators)):
            base_estimator_path = os.path.join(
                base_estimators_base_path,
                str(model_index),
            )
            save_to_store(base_estimator_path, self.base_estimators[model_index])

    def from_dir(self, directory: str) -> None:
        """Deserialises the backend model object stored in a given directory.

        As ensembles have a meta estimator and a list of base estimators, each one is
        saved to a subdirectory, in the format:
        `directory/
            meta_estimator/
                ...
            base_estimators/
                0/
                    ...
                1/
                    ...
                ...
        """

        # The expected model binary
        meta_estimator_path = os.path.join(directory, "meta_estimator")
        base_estimators_base_path = os.path.join(directory, "base_estimators")

        self.meta_estimator = load_from_store(meta_estimator_path)

        base_estimators = []
        for model_index in range(len(self.base_estimators)):
            base_estimator_path = os.path.join(
                base_estimators_base_path,
                str(model_index),
            )
            base_estimators.append(load_from_store(base_estimator_path))

        self.base_estimators = base_estimators



from dataclasses import field
from typing import Any

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)
from molflux.modelzoo.models.sklearn.sklearn_pipeline._utils import (
    StepConfigsT,
    build_pipeline,
)
from molflux.modelzoo.utils import format_wrapped_model_tag

try:
    from sklearn.base import is_classifier
    from sklearn.pipeline import Pipeline

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from e


_DESCRIPTION = """
This is an implementation of an SKLearn Pipeline Classifier. It takes in a list of dictionaries,
each defining a step in the pipeline. This has to conform to the official user guide referenced below. The last
step must be an estimator classifier.

By default, this returns a Pipeline with a StandardScaler before an SKLearn Random Forest Classifier.

Reference: https://scikit-learn.org/stable/modules/compose.html#pipeline
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
step_configs: Optional[list] = None
    List defining each step of an SKLearn Pipeline. The format of the list must follow
    this format:
    [
        {
            "class_name": "sklearn.some_module.ClassName",
            "hyperparameters": {
                "first_hyperparam_for_ClassName": 500,
                "second_hp_for_ClassName": "auto",
                "learning_rate": 0.1,
                ...
            },
        },
        {
            "class_name": ...
            "hyperparameters": ...
        },
        ...
    ]

    It is a list of dictionaries, with a class name and hyperparameters for each step. The last one
    is expected to be a classifier.
"""


DEFAULT_CONFIG: StepConfigsT = [
    {
        "class_name": "sklearn.preprocessing.StandardScaler",
        "hyperparameters": None,
    },
    {
        "class_name": "sklearn.ensemble.RandomForestClassifier",
        "hyperparameters": {
            "n_estimators": 500,
            "random_state": 0,
        },
    },
]


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class SklearnPipelineClassifierConfig(ModelConfig):
    step_configs: StepConfigsT = field(
        default_factory=lambda: DEFAULT_CONFIG.copy(),
    )


class SklearnPipelineClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[SklearnPipelineClassifierConfig],
):
    def __init__(self, tag: str | None = None, **config_kwargs: Any) -> None:
        super().__init__(tag=tag, **config_kwargs)
        if tag is None:
            self.info.tag = format_wrapped_model_tag(
                self.info.tag,
                [self.model_config.step_configs[-1].get("class_name", "")],
            )

    @property
    def _config_builder(self) -> type[SklearnPipelineClassifierConfig]:
        return SklearnPipelineClassifierConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> Pipeline:
        config = self.model_config
        pipeline = build_pipeline(config.step_configs)

        # assert that the last estimator of this pipeline is a classifier
        pipeline_estimator = pipeline[-1]

        if not is_classifier(pipeline_estimator):
            raise RuntimeError(
                f"The last step of the pipeline is expected to be a classifier. Found instead "
                f"{pipeline_estimator}",
            )

        return pipeline



from collections.abc import Iterable, Iterator
from typing import Any

import sklearn.model_selection

from molflux.splits.bases import SplittingStrategyBase
from molflux.splits.info import SplittingStrategyInfo
from molflux.splits.typing import ArrayLike, SplitIndices, Splittable

_DESCRIPTION = """
Time Series cross-validator

Provides train/validation indices to split time series data samples
that are observed at fixed time intervals, in train/validation sets.
In each split, test indices must be higher than before, and thus shuffling
in cross validator is inappropriate.

This cross-validation object is a variation of :class:`KFold`.
In the kth split, it returns first k folds as train set and the
(k+1)th fold as test set.

Note that unlike standard cross-validation methods, successive
training sets are supersets of those that come before them.
"""


class TimeSeriesSplit(SplittingStrategyBase):
    def _info(self) -> SplittingStrategyInfo:
        return SplittingStrategyInfo(
            description=_DESCRIPTION,
        )

    def _split(
        self,
        dataset: Splittable,
        y: ArrayLike | None = None,
        groups: ArrayLike | None = None,
        *,
        n_splits: int = 2,
        max_train_size: int | None = None,
        test_size: int | None = None,
        gap: int | None = 0,
        **kwargs: Any,
    ) -> Iterator[SplitIndices]:
        """
        Args:
            dataset: The data to be split.
            y (optional): The target variable for supervised learning problems.
            groups (optional): Group labels for the samples used while splitting the dataset.
            n_splits (optional): The number of splits to generate. Must be greater than one. Defaults to 2.
            max_train_size (optional): The maximum size for a single training set.
            test_size (optional): Used to limit the size of the test set. Defaults to
                ``n_samples // (n_splits + 1)``, which is the maximum allowed value
                with ``gap=0``.
            gap (optional): Number of samples to exclude from the end of each train set
                before the test set.
        """
        # Wrap scikit-learn implementation
        cv = sklearn.model_selection.TimeSeriesSplit(
            n_splits=n_splits,
            max_train_size=max_train_size,
            test_size=test_size,
            gap=gap,
        )

        for indices in cv.split(dataset, y=y, groups=groups):
            train_indices, validation_indices = indices
            test_indices: Iterable[int] = []  # no holdout in cross-validation
            yield train_indices, validation_indices, test_indices



"""Compute average precision (AP) from prediction scores."""

import logging
from typing import Any

import evaluate
from sklearn.metrics import average_precision_score

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
AP summarizes a precision-recall curve as the weighted mean of precisions achieved at each threshold,
with the increase in recall from the previous threshold used as the weight. This implementation is not
interpolated and is different from computing the area under the precision-recall curve with the trapezoidal
rule, which uses linear interpolation and can be too optimistic.

Note: this implementation is restricted to the binary classification task or multilabel classification task.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Target prediction scores.
        Target scores, can either be probability estimates of the positive class,
        confidence values, or non-thresholded measure of decisions.

    references:
        True binary labels or binary label indicators.

    average: {`micro`, `samples`, `weighted`, `macro`} or None, default=`macro`
        If None, the scores for each class are returned. Otherwise, this determines the type
        of averaging performed on the data:

        'micro':
            Calculate metrics globally by considering each molflux of the label indicator matrix as a label.

        'macro':
            Calculate metrics for each label, and find their unweighted mean. This does not take label imbalance
            into account.

        'weighted':
            Calculate metrics for each label, and find their average, weighted by support (the number of true
            instances for each label).

        'samples':
            Calculate metrics for each instance, and find their average.

        Will be ignored when y_true is binary.

    pos_label: int or str, default=1
        The label of the positive class. Only applied to binary y_true. For multilabel-indicator y_true,
        pos_label is fixed to 1.

    sample_weight: array-like of shape (n_samples,), default=None
        Sample weights.

Returns:
    average_precision: The average precision score.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("average_precision")
    >>> references = [1, 0, 1, 1, 0, 0]
    >>> predictions = [0.5, 0.2, 0.99, 0.3, 0.1, 0.7]
    >>> metric.compute(references=references, predictions=predictions)
    {'average_precision': 0.80...}
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("average_precision", config_name="multilabel")
    >>> references = [[1, 1, 0],
    ...         [1, 1, 0],
    ...         [0, 1, 0],
    ...         [0, 0, 1],
    ...         [0, 1, 1],
    ...         [1, 0, 1]]
    >>> predictions = [[0.3, 0.5, 0.2],
    ...                 [0.7, 0.2, 0.1],
    ...                 [0.005, 0.99, 0.005],
    ...                 [0.2, 0.3, 0.5],
    ...                 [0.1, 0.1, 0.8],
    ...                 [0.1, 0.7, 0.2]]
    >>> results = metric.compute(references=references, predictions=predictions, average=None)
    >>> print([round(res, 2) for res in results['average_precision']])
    [0.87, 0.73, 0.92]
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class AveragePrecision(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("int32")),
                }
                if self.config_name == "multilabel"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("int32"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.average_precision_score.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: list[float] | None = None,
        average: str = "macro",
        pos_label: int | str = 1,
        **kwargs: Any,
    ) -> MetricResult:
        score = average_precision_score(
            y_true=references,
            y_score=predictions,
            average=average,
            sample_weight=sample_weight,
            pos_label=pos_label,
        )
        return {self.tag: score}



from collections.abc import Callable
from typing import Any, Literal

import numpy as np
import scipy.stats as st
from pydantic.v1 import dataclasses

import datasets
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import (
    ModelBase,
    ModelConfig,
    PredictionIntervalMixin,
    SamplingMixin,
    StandardDeviationMixin,
)
from molflux.modelzoo.typing import Features, PredictionResult

_DESCRIPTION = """
A core model that predicts the average (median/mean) of the features.
A Gaussian mixture model approximation can also be used by providing std associated with each feature.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
average_method: Literal, default = "median". Method by which to summarise the average of the dataset.
    Available options are median or mean. median is preferred since it is robust to outliers.
deviation_method: Literal, default = "mad". Method by which to summarise the deviation of the dataset.
    Available options are std or mad. Note that mad is the median absolute deviation which is a version
    of standard deviation robust to outlier values.
x_std_features: Features, default = None. List of strings providing a subset of the features that
    are standard deviations associated with other features for use in a Gaussian mixture model.
"""

# mapping from method strings to fns
_AVERAGE_METHOD_DICT: dict[str, Callable] = {
    "median": np.nanmedian,
    "mean": np.nanmean,
}
_DEVIATION_METHOD_DICT: dict[str, Callable] = {
    "std": np.nanstd,
    "mad": st.median_abs_deviation,
}


class Config:
    arbitrary_types_allowed = True


@dataclasses.dataclass
class AverageModel:
    average_fn: Callable
    deviation_fn: Callable


@dataclasses.dataclass(config=Config)
class AverageFeaturesRegressorConfig(ModelConfig):
    average_method: Literal["median", "mean"] = "median"
    deviation_method: Literal["std", "mad"] = "mad"
    x_std_features: Features | None = None


class AverageFeaturesRegressor(
    PredictionIntervalMixin,
    StandardDeviationMixin,
    SamplingMixin,
    ModelBase[AverageFeaturesRegressorConfig],
):
    @property
    def _config_builder(self) -> type[AverageFeaturesRegressorConfig]:
        return AverageFeaturesRegressorConfig

    def _info(self) -> ModelInfo:
        """Initialises the ModelInfo object."""
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> Any:
        if self.model_config.x_std_features is not None:
            assert (
                len(self.x_features)
                == 2
                * len(
                    self.model_config.x_std_features,
                )
            ), "When providing stds (``x_std_features``), these should match to the means with both of these provided in ``x_features``"
        self.model = AverageModel(
            average_fn=_AVERAGE_METHOD_DICT[self.model_config.average_method],
            deviation_fn=_DEVIATION_METHOD_DICT[self.model_config.deviation_method],
        )

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> PredictionResult:
        display_names = self._predict_display_names
        if not len(data):
            return {display_name: [] for display_name in display_names}

        data = data.with_format("np")  # easier for processing of missing None values
        assert isinstance(
            self.model,
            AverageModel,
        ), "Ensure ``average_features_regressor`` model has been trained"
        x_features = (
            list(set(self.x_features) - set(self.model_config.x_std_features))
            if self.model_config.x_std_features is not None
            else self.x_features
        )
        params_dict = {
            "mu": [
                self.model.average_fn(list(row.values()))
                for row in data.select_columns(x_features)
            ],
        }

        return {display_name: params_dict["mu"] for display_name in display_names}

    def _predict_with_std(
        self,
        data: datasets.Dataset,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_std_display_names,
        ) = self._predict_with_std_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_std_display_names
            }

        data = data.with_format("np")  # easier for processing of missing None values
        assert isinstance(
            self.model,
            AverageModel,
        ), "Ensure ``average_features_regressor`` model has been trained"

        x_features = (
            list(set(self.x_features) - set(self.model_config.x_std_features))
            if self.model_config.x_std_features is not None
            else self.x_features
        )
        filtered_data_x = data.select_columns(x_features)
        params_dict = {
            "mu": [
                self.model.average_fn(list(row.values())) for row in filtered_data_x
            ],
        }
        if self.model_config.x_std_features is not None:
            # Use a Gaussian to approximate a mixture of Gaussians from each of the individual models
            filtered_data_x_std = data.select_columns(self.model_config.x_std_features)
            sigma_sq = []
            for row_x, row_x_std, mu_star in zip(
                filtered_data_x,
                filtered_data_x_std,
                params_dict["mu"],
                strict=False,
            ):
                sigma_sq.append(
                    self.model.average_fn(
                        [
                            x**2 + y**2
                            for x, y in zip(
                                row_x.values(),
                                row_x_std.values(),
                                strict=False,
                            )
                        ],
                    )
                    - mu_star**2,
                )
            params_dict["sigma"] = [np.sqrt(s_sq) for s_sq in sigma_sq]
        else:
            params_dict["sigma"] = [
                self.model.deviation_fn(list(row.values())) for row in filtered_data_x
            ]

        return {
            display_name: params_dict["mu"] for display_name in prediction_display_names
        }, {
            display_name: params_dict["sigma"]
            for display_name in prediction_std_display_names
        }

    def _predict_with_prediction_interval(
        self,
        data: datasets.Dataset,
        confidence: float,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        (
            prediction_display_names,
            prediction_interval_display_names,
        ) = self._predict_with_prediction_interval_display_names

        if not len(data):
            return {display_name: [] for display_name in prediction_display_names}, {
                display_name: [] for display_name in prediction_interval_display_names
            }

        # get the result dictionary of means and standard deviations
        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        # for each column, map the mean and standard deviation to a prediction interval
        prediction_results: PredictionResult = {}
        prediction_prediction_interval_results: PredictionResult = {}
        for prediction_display_name, prediction_interval_display_name, mean, std in zip(
            prediction_display_names,
            prediction_interval_display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            # compute the prediction interval
            lower_bound, upper_bound = st.norm.interval(
                confidence,
                loc=mean,
                scale=std,
            )

            # where nans occur (ex. from a 0 standard deviation), use the mean instead
            lower_bound = np.where(~np.isnan(lower_bound), lower_bound, mean)
            upper_bound = np.where(~np.isnan(upper_bound), upper_bound, mean)

            prediction_results[prediction_display_name] = mean
            prediction_prediction_interval_results[prediction_interval_display_name] = (
                list(zip(lower_bound, upper_bound, strict=False))
            )

        return prediction_results, prediction_prediction_interval_results

    def _sample(
        self,
        data: datasets.Dataset,
        n_samples: int,
        **kwargs: Any,
    ) -> PredictionResult:
        display_names = self._sample_display_names

        if not len(data):
            return {display_name: [] for display_name in display_names}

        prediction_mean_results, prediction_std_results = self._predict_with_std(data)

        prediction_results: PredictionResult = {}
        for display_name, means, stds in zip(
            display_names,
            prediction_mean_results.values(),
            prediction_std_results.values(),
            strict=False,
        ):
            samples = np.random.normal(means, stds, (n_samples, len(means))).T
            prediction_results[display_name] = samples.tolist()

        return prediction_results



import pytest

from molflux.metrics.load import load_suite
from molflux.metrics.metric import Metrics
from molflux.metrics.suites.catalogue import list_suites

suite_name = "regression"


def test_in_catalogue():
    """That the suite is in the catalogue."""
    suites = list_suites()
    assert suite_name in suites


def test_loads():
    """That the suite can be loaded without errors."""
    assert load_suite(name=suite_name)


def test_is_metrics():
    """That the suite returns a metrics collection."""
    suite = load_suite(name=suite_name)
    assert isinstance(suite, Metrics)


def test_results_have_all_the_metrics_in():
    """Check that results have the regression metrics."""
    ground_truth = [0, 0.3, 0.5, 0.8, 1]
    preds = [0.1, 0.35, 0.45, 0.68, 1.2]
    suite = load_suite(name=suite_name)
    result = suite.compute(references=ground_truth, predictions=preds)
    all_metrics = [metric.name for metric in suite]
    assert set(result.keys()) - set(all_metrics) == {
        "spearman::p_value",
        "spearman::correlation",
        "pearson::p_value",
        "pearson::correlation",
    }
    assert set(all_metrics) - set(result.keys()) == {"pearson", "spearman"}


def test_compute_works():
    """Check that computations work for a few of the
    metrics in the regression suite."""
    ground_truth = [0, 0.3, 0.5, 0.8, 1]
    preds = [0.1, 0.35, 0.45, 0.68, 1.2]
    suite = load_suite(name=suite_name)
    result = suite.compute(references=ground_truth, predictions=preds)
    assert pytest.approx(result["r2"], 0.01) == 0.88949
    assert pytest.approx(result["mean_absolute_error"], 0.01) == 0.10400



from typing import Literal

import datasets

URL_DICT: dict[str, dict[str, str]] = {
    "wB97X/631Gd": {
        "url": "https://zenodo.org/records/10108942/files/ANI-2x-wB97X-631Gd.tar.gz",
        "path": "final_h5/ANI-2x-wB97X-631Gd.h5",
    },
    "wB97X/def2TZVPP": {
        "url": "https://zenodo.org/records/10108942/files/ANI-2x-wB97X-def2TZVPP.tar.gz",
        "path": "final_h5/ANI-2x-wB97X-def2TZVPP.h5",
    },
    "wB97MD3BJ/def2TZVPP": {
        "url": "https://zenodo.org/records/10108942/files/ANI-2x-wB97MD3BJ-def2TZVPP.tar.gz",
        "path": "final_h5/ANI-2x-wB97MD3BJ-def2TZVPP.h5",
    },
    "wB97MV/def2TZVPP": {
        "url": "https://zenodo.org/records/10108942/files/ANI-2x-wB97MV-def2TZVPP.tar.gz",
        "path": "final_h5/ANI-2x-wB97MV-def2TZVPP.h5",
    },
    "B973c/def2mTZVP": {
        "url": "https://zenodo.org/records/10108942/files/ANI-2x-B973c-def2mTZVP.tar.gz",
        "path": "final_h5/ANI-2x-B973c-def2mTZVP.h5",
    },
}

LEVEL_OF_THEORY = Literal[
    "wB97X/631Gd",
    "wB97X/def2TZVPP",
    "wB97MD3BJ/def2TZVPP",
    "wB97MV/def2TZVPP",
    "B973c/def2mTZVP",
]

FEATURES = {
    "wB97X/631Gd": {
        "mol_bytes": datasets.Value("binary"),
        "energies": datasets.Value("float64"),
        "forces": datasets.Sequence(
            feature=datasets.Sequence(
                feature=datasets.Value(dtype="float64", id=None),
                length=-1,
                id=None,
            ),
            length=-1,
            id=None,
        ),
    },
    "wB97X/def2TZVPP": {
        "mol_bytes": datasets.Value("binary"),
        "energies": datasets.Value("float64"),
        "dipoles": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "mbis_atomic_charges": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "mbis_atomic_dipole_magnitudes": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "mbis_atomic_volumes": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "mbis_atomic_octupole_magnitudes": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "mbis_atomic_quadrupole_magnitudes": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
    },
    "wB97MD3BJ/def2TZVPP": {
        "mol_bytes": datasets.Value("binary"),
        "energies": datasets.Value("float64"),
        "dipoles": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "wB97M_def2-TZVPP.scf-energies": datasets.Value("float64"),
        "D3.energy-corrections": datasets.Value("float64"),
    },
    "wB97MV/def2TZVPP": {
        "mol_bytes": datasets.Value("binary"),
        "energies": datasets.Value("float64"),
        "dipoles": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "wB97M_def2-TZVPP.scf-energies": datasets.Value("float64"),
        "VV10.energy-corrections": datasets.Value("float64"),
    },
    "B973c/def2mTZVP": {
        "mol_bytes": datasets.Value("binary"),
        "energies": datasets.Value("float64"),
        "forces": datasets.Sequence(
            feature=datasets.Sequence(
                feature=datasets.Value(dtype="float64", id=None),
                length=-1,
                id=None,
            ),
            length=-1,
            id=None,
        ),
        "dipole": datasets.Sequence(
            datasets.Value("float64"),
            length=-1,
        ),
        "D3.energy-corrections": datasets.Value("float64"),
        "D3.force-corrections": datasets.Sequence(
            feature=datasets.Sequence(
                feature=datasets.Value(dtype="float64", id=None),
                length=-1,
                id=None,
            ),
            length=-1,
            id=None,
        ),
    },
}



import numpy as np
import pytest

from molflux.splits.catalogue import list_splitting_strategies
from molflux.splits.load import load_splitting_strategy
from molflux.splits.strategy import SplittingStrategy

strategy_name = "stratified_k_fold"


@pytest.fixture(scope="module")
def fixture_test_strategy():
    return load_splitting_strategy(strategy_name)


@pytest.fixture(scope="module")
def fixture_sample_inputs():
    n = 100
    dataset = np.random.rand(n)
    classes = np.random.choice([0, 1, 2], size=n)
    return dataset, classes


def test_is_in_catalogue():
    """That the strategy is registered in the catalogue."""
    catalogue = list_splitting_strategies()
    all_strategy_names = [name for names in catalogue.values() for name in names]
    assert strategy_name in all_strategy_names


def test_implements_protocol(fixture_test_strategy):
    """That the strategy implements the protocol."""
    strategy = fixture_test_strategy
    assert isinstance(strategy, SplittingStrategy)


def test_requires_classes(fixture_test_strategy):
    """That classes need to be provided for stratification strategy."""
    dataset = [0, 1, 2, 3, 4, 5]
    strategy = fixture_test_strategy
    with pytest.raises(ValueError, match="The 'y' parameter should not be None"):
        next(strategy.split(dataset))


def test_accepts_integer_class_labels(fixture_test_strategy):
    """That the splitting strategy accepts integer class labels."""
    dataset = np.random.rand(10)
    classes = np.random.choice([0, 1, 2], size=10)
    strategy = fixture_test_strategy
    next(strategy.split(dataset, y=classes))
    assert True


def test_accepts_string_class_labels(fixture_test_strategy):
    """That the splitting strategy accepts string class labels."""
    dataset = np.random.rand(10)
    classes = np.random.choice(["cat", "dog", "snake"], size=10)
    strategy = fixture_test_strategy
    next(strategy.split(dataset, y=classes))
    assert True


def test_yields_more_than_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That the splitting strategy only yields several folds."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes)
    assert len(list(indices)) > 1


def test_cannot_set_only_one_fold(fixture_sample_inputs, fixture_test_strategy):
    """That cannot perform cross validation with only one fold."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    with pytest.raises(ValueError, match="got n_splits=1"):
        next(strategy.split(dataset, y=classes, n_splits=1))


def test_raises_on_inconsistent_shuffle_and_seed(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    """That an error is raised if seed is set but shuffle is not requested."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    with pytest.raises(
        ValueError,
        match="Setting a random_state has no effect since shuffle is False",
    ):
        next(strategy.split(dataset, y=classes, seed=123, shuffle=False))


def test_split_fractions(fixture_sample_inputs, fixture_test_strategy):
    """That the k folds are of expected sizes."""
    dataset, classes = fixture_sample_inputs
    n_samples = len(dataset)
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes, n_splits=4)
    for train_indices, validation_indices, test_indices in indices:
        assert len(list(train_indices)) == 0.75 * n_samples
        assert len(list(validation_indices)) == 0.25 * n_samples
        assert len(list(test_indices)) == 0


def test_splits_are_disjoint(fixture_sample_inputs, fixture_test_strategy):
    """That data is spread across splits without overlap."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy
    indices = strategy.split(dataset, y=classes)
    for train_indices, validation_indices, _ in indices:
        assert set(train_indices).isdisjoint(set(validation_indices))


def test_unshuffled_by_default(fixture_test_strategy):
    """That by default the data is not shuffled before splitting."""
    dataset = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    classes = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
    strategy = fixture_test_strategy

    indices = strategy.split(dataset, y=classes, n_splits=2)
    train_indices1, validation_indices1, _ = next(indices)
    train_indices2, validation_indices2, _ = next(indices)

    # We expect KFold to assign data to the validation set starting from the top
    # The expected results is:
    # [5 6 7 8 9] [0 1 2 3 4] ()
    # [0 1 2 3 4] [5 6 7 8 9] ()

    # fold 1
    assert all(i == j for i, j in zip(train_indices1, [5, 6, 7, 8, 9], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices1, [0, 1, 2, 3, 4], strict=False)
    )

    # fold 2
    assert all(i == j for i, j in zip(train_indices2, [0, 1, 2, 3, 4], strict=False))
    assert all(
        i == j for i, j in zip(validation_indices2, [5, 6, 7, 8, 9], strict=False)
    )


def test_deterministic_by_default(fixture_sample_inputs, fixture_test_strategy):
    """That by default repeated splits are deterministic (as not shuffled)."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, y=classes, n_splits=2)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, y=classes, n_splits=2)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we get the same results across folds
    assert all(i == j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert all(i == j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )


def test_non_deterministic_if_shuffled(fixture_sample_inputs, fixture_test_strategy):
    """That repeated splits are not deterministic if dataset is shuffled."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, y=classes, n_splits=2, shuffle=True)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, y=classes, n_splits=2, shuffle=True)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we do not get the same results across folds
    assert any(i != j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert any(
        i != j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert any(i != j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert any(
        i != j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )


def test_deterministic_if_shuffled_and_seed_set(
    fixture_sample_inputs,
    fixture_test_strategy,
):
    """That split results are deterministic if dataset is shuffled but a seed is set."""
    dataset, classes = fixture_sample_inputs
    strategy = fixture_test_strategy

    # Do one iteration of k-fold splits
    indices_a = strategy.split(dataset, y=classes, n_splits=2, shuffle=True, seed=123)
    train_indices_a1, validation_indices_a1, _ = next(indices_a)
    train_indices_a2, validation_indices_a2, _ = next(indices_a)

    # And again...
    indices_b = strategy.split(dataset, y=classes, n_splits=2, shuffle=True, seed=123)
    train_indices_b1, validation_indices_b1, _ = next(indices_b)
    train_indices_b2, validation_indices_b2, _ = next(indices_b)

    # Check that we didn't get the same results across folds
    assert all(i == j for i, j in zip(train_indices_a1, train_indices_b1, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a1, validation_indices_b1, strict=False)
    )
    assert all(i == j for i, j in zip(train_indices_a2, train_indices_b2, strict=False))
    assert all(
        i == j
        for i, j in zip(validation_indices_a2, validation_indices_b2, strict=False)
    )



import logging
from typing import Any, TypeVar

import molflux.datasets
import molflux.features
from molflux.core.featurisation.metadata import (
    FeaturisationMetadataV1,
    FeaturisationMetadataV2,
    fetch_model_featurisation_metadata,
    parse_featurisation_metadata,
)
from molflux.core.typing import DatasetType

logger = logging.getLogger(__name__)


# Generic type; used to indicate functions are Dataset->Dataset, DatasetDict->DatasetDict
DatasetT = TypeVar("DatasetT", bound=DatasetType)


def replay_dataset_featurisation(
    inputs: DatasetType,
    model_path: str,
    **map_kwargs: Any,
) -> DatasetType:
    featurisation_metadata = fetch_model_featurisation_metadata(model_path=model_path)
    return featurise_dataset(
        inputs=inputs,
        featurisation_metadata=featurisation_metadata,
        **map_kwargs,
    )


def featurise_dataset(
    inputs: DatasetType,
    featurisation_metadata: dict[str, Any],
    **map_kwargs: Any,
) -> DatasetType:
    featurisation_metadata, version = parse_featurisation_metadata(
        featurisation_metadata,
    )

    if version == 1:
        return _featurise_dataset_v1(
            dataset=inputs,
            featurisation_metadata=featurisation_metadata,
            **map_kwargs,
        )
    if version == 2:
        return _featurise_dataset_v2(
            dataset=inputs,
            featurisation_metadata=featurisation_metadata,
            **map_kwargs,
        )
    else:
        raise NotImplementedError(
            f"No featuriser implemented for featurisation metadata version: {version!r}",
        )


def _featurise_dataset_v1(
    dataset: DatasetType,
    featurisation_metadata: dict[str, Any],
    **map_kwargs: Any,
) -> DatasetType:
    """Featurises a dataset from a V1 featurisation metadata schema."""
    featurisation_metadata_obj = FeaturisationMetadataV1(**featurisation_metadata)
    featurisation_config = featurisation_metadata_obj.config
    for column_featurisation_config in featurisation_config:
        column = column_featurisation_config.column
        representation_configs = column_featurisation_config.representations

        # We allow passing single string display_names as convenience method
        # but we need to turn them into a list to match canonical form expected
        # by molflux.datasets.featurise_dataset()
        display_names = [config.as_ for config in representation_configs]
        canonical_display_names = [
            names if isinstance(names, list) else [names] for names in display_names
        ]

        representation_configs_as_dicts = [
            {"name": config.name, "config": config.config, "presets": config.presets}
            for config in representation_configs
        ]

        representations = molflux.features.load_from_dicts(
            representation_configs_as_dicts,
        )
        dataset = molflux.datasets.featurise_dataset(
            dataset,
            column=[column],
            representations=representations,
            display_names=canonical_display_names,  # type: ignore[arg-type]
            **map_kwargs,
        )
    return dataset


def _featurise_dataset_v2(
    dataset: DatasetT,
    featurisation_metadata: dict[str, Any],
    **map_kwargs: Any,
) -> DatasetT:
    """Featurises a dataset from a V2 featurisation metadata schema."""
    featurisation_metadata_obj = FeaturisationMetadataV2(**featurisation_metadata)
    featurisation_config = featurisation_metadata_obj.config
    for column_featurisation_config in featurisation_config:
        columns = column_featurisation_config.columns
        representation_configs = column_featurisation_config.representations

        # We allow passing single string display_names as convenience method
        # but we need to turn them into a list to match canonical form expected
        # by molflux.datasets.featurise_dataset()
        display_names = [config.as_ for config in representation_configs]
        canonical_display_names = [
            names if isinstance(names, list) else [names] for names in display_names
        ]

        representation_configs_as_dicts = [
            {"name": config.name, "config": config.config, "presets": config.presets}
            for config in representation_configs
        ]

        representations = molflux.features.load_from_dicts(
            representation_configs_as_dicts,
        )
        dataset = molflux.datasets.featurise_dataset(
            dataset,
            column=columns,
            representations=representations,
            display_names=canonical_display_names,  # type: ignore[arg-type]
            **map_kwargs,
        )
    return dataset



"""Spearman correlation coefficient score function."""

import logging
from typing import Any, Literal

import evaluate
import scipy.stats

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
Calculate a Spearman correlation coefficient with associated p-value.

The Spearman rank-order correlation coefficient is a nonparametric measure of
the monotonicity of the relationship between two datasets. Unlike the Pearson
correlation, the Spearman correlation does not assume that both datasets are
normally distributed. Like other correlation coefficients, this one varies
between -1 and +1 with 0 implying no correlation. Correlations of -1 or +1 imply
an exact monotonic relationship. Positive correlations imply that as x increases,
so does y. Negative correlations imply that as x increases, y decreases.

The p-value roughly indicates the probability of an uncorrelated system producing
datasets that have a Spearman correlation at least as extreme as the one computed
from these datasets. The p-values are not entirely reliable but are probably
reasonable for datasets larger than 500 or so.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    axis (optional): The axis to evaluate the correlation against.
        If axis=0 (default), then each column represents a variable, with
        observations in the rows. If axis=1, the relationship is transposed:
        each row represents a variable, while the columns contain observations.
        If axis=None, then both arrays will be raveled.
    nan_policy (optional): Defines aggregating of multiple output scores.
        The following options are available (default is 'propagate'):

        * 'propagate': returns nan
        * 'raise': throws an error
        * 'omit': performs the calculations ignoring nan values

    alternative (optional): Defines the alternative hypothesis. Default is 'two-sided'.
        The following options are available:

        * 'two-sided': the correlation is nonzero
        * 'less': the correlation is negative (less than zero)
        * 'greater':  the correlation is positive (greater than zero)

Returns:
    The Spearman correlation matrix or correlation coefficient (if only 2
    variables are given as parameters. Correlation matrix is square with
    length equal to total number of variables (columns or rows) in ``predictions``
    and ``references`` combined.

    The p-value for a hypothesis test whose null hypotheisis
    is that two sets of data are uncorrelated. See `alternative` above
    for alternative hypotheses. `pvalue` has the same
    shape as `correlation`.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("spearman")
    >>> predictions = [1, 2, 3, 4, 5]
    >>> references = [5, 6, 7, 8, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'spearman::correlation': 0.82078..., 'spearman::p_value': 0.08858...}
"""

_CITATION = """\
Zwillinger, D. and Kokoska, S. (2000). CRC Standard Probability and Statistics
Tables and Formulae. Chapman & Hall: New York. 2000. Section 14.7
"""

NanPolicy = Literal["propagate", "raise", "omit"]
Alternative = Literal["two-sided", "less", "greater"]


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class Spearman(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.spearmanr.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        axis: int | None = 0,
        nan_policy: NanPolicy = "propagate",
        alternative: Alternative = "two-sided",
        **kwargs: Any,
    ) -> MetricResult:
        correlation, p_value = scipy.stats.spearmanr(
            a=predictions,
            b=references,
            axis=axis,
            nan_policy=nan_policy,
            alternative=alternative,
        )
        return {
            f"{self.tag}::correlation": correlation,
            f"{self.tag}::p_value": p_value,
        }



from collections.abc import Iterable

from pydantic.v1 import dataclasses

from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.sklearn import (
    SKLearnClassificationMixin,
    SKLearnModelBase,
)

try:
    from molflux.modelzoo.models.sklearn.sklearn_discrete_nb.bayes import (  # type: ignore
        CorrectedNB as SKCorrectedNB,
    )
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("sklearn", e) from None


_DESCRIPTION = """
A Naive Bayes classifier with corrected probabilities
based on the paper https://pubs.acs.org/doi/10.1021/jm0303195.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
alpha : float or array-like of shape (n_features,), default=1.0
    Additive (Laplace/Lidstone) smoothing parameter
    (set alpha=0 for no smoothing).

fit_prior : bool, default=True
    Whether to learn class prior probabilities or not.
    If false, a uniform prior will be used.

class_prior : array-like of shape (n_classes,), default=None
    Prior probabilities of the classes. If specified, the priors are not
    adjusted according to the data.
"""


class Config:
    arbitrary_types_allowed = True
    extra = "forbid"


@dataclasses.dataclass(config=Config)
class CorrectedNBConfig(ModelConfig):
    alpha: float = 1
    fit_prior: bool = True
    class_prior: Iterable | None = None


class CorrectedNBClassifier(
    SKLearnClassificationMixin,
    SKLearnModelBase[CorrectedNBConfig],
):
    @property
    def _config_builder(self) -> type[CorrectedNBConfig]:
        return CorrectedNBConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    def _instantiate_model(self) -> SKCorrectedNB:
        config = self.model_config
        return SKCorrectedNB(
            alpha=config.alpha,
            fit_prior=config.fit_prior,
            class_prior=config.class_prior,
        )



from unittest.mock import patch

import pytest
from openeye.oechem import OEMol, OEReadMolFromBytes

import datasets
from molflux.datasets import load_dataset
from molflux.datasets.builders.ani2x.ani2x import ANI2X
from molflux.datasets.builders.ani2x.ani2x_configs import FEATURES
from molflux.datasets.catalogue import list_datasets

dataset_name = "ani2x"
backend_name = "openeye"
levels_of_theory = [
    "wB97X/631Gd",
    "wB97X/def2TZVPP",
    "wB97MD3BJ/def2TZVPP",
    "wB97MV/def2TZVPP",
    "B973c/def2mTZVP",
]


@pytest.fixture()
def _fixture_import_main_class_patch():
    """Mocks out the import_main_class from datasets'.

    This function is used when generating a dataset builder in the `datasets` class.
    It is done by copying the file path of a builder script into a temporary folder
    under the huggingface cache directory, and loading the Builder class dynamically
    at runtime. This means that any test for a builder will instead be tested on a
    copy of the builder script, not showing up as covered in any coverage run.
    Instead, we redirect the builder class import to the actual class in datasets.
    """
    with patch("datasets.load.import_main_class") as mock:
        mock.return_value = ANI2X
        yield


pytestmark = pytest.mark.usefixtures("_fixture_import_main_class_patch")


@pytest.fixture(scope="module")
def fixture_path_to_mock_data(fixture_path_to_this_test):
    """The path to the directory containing mock data for this builder."""
    here = fixture_path_to_this_test
    return here / "mock_data"


@pytest.fixture()
def _fixture_mocked_dataset_asset(fixture_path_to_mock_data):
    """Mocks out the DownloadManager used in 'datasets.load_dataset()'.

    The DownloadManager is used in the builder's '_split_generator()`, which
    is the function that actually downloads all the data for a given dataset.
    In this way we can avoid having to download the real datasets from AWS S3,
    and return paths to local assets.
    """
    with patch("datasets.builder.DownloadManager.download_and_extract") as mock:
        mock.return_value = str(fixture_path_to_mock_data)
        yield


def test_in_catalogue():
    """That the builder is registered in the catalogue."""
    catalogue = list_datasets()
    all_names = [name for names in catalogue.values() for name in names]
    assert dataset_name in all_names


@pytest.mark.parametrize(
    ("level_of_theory"),
    levels_of_theory,
)
@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_returns_huggingface_dataset(level_of_theory):
    """That the built dataset is a huggingface dataset."""
    dataset = load_dataset(dataset_name, backend_name, level_of_theory=level_of_theory)
    assert isinstance(dataset, datasets.Dataset)


@pytest.mark.parametrize(
    ("level_of_theory"),
    levels_of_theory,
)
@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_dataset_has_correct_num_rows(level_of_theory):
    """That the built dataset has correct num rows."""

    dataset = load_dataset(dataset_name, backend_name, level_of_theory=level_of_theory)

    assert set(dataset.column_names) == set(FEATURES[level_of_theory].keys())
    assert len(dataset) == 10


@pytest.mark.parametrize(
    ("level_of_theory"),
    levels_of_theory,
)
@pytest.mark.usefixtures("_fixture_mocked_dataset_asset")
def test_dataset_is_readable_with_openeye(level_of_theory):
    """That openeye can read the mol bytes"""

    dataset = load_dataset(dataset_name, backend_name, level_of_theory=level_of_theory)

    for mol_bytes in dataset["mol_bytes"]:
        mol = OEMol()
        OEReadMolFromBytes(mol, ".oeb", mol_bytes)



import logging
import warnings
from typing import Any, Union

import pyarrow as pa
from more_itertools import zip_broadcast

import datasets
from molflux.datasets.interfaces import Representation, Representations
from molflux.datasets.typing import DatasetType, DisplayNames

_DEFAULT_DISPLAY_NAMES_TEMPLATE = "{source_column}::{feature_name}"

FreeformDisplayNames = Union[
    str | None,
    list[str | None | list[str | None]],
]


logger = logging.getLogger(__name__)


def _to_canonical_display_names(
    display_names: FreeformDisplayNames,
    representations: Representations,
) -> DisplayNames:
    """Parses free-form display names provided by users into their canonical form.

    Display names in canonical form are represented as a nested list of display
    names, with one (list) entry for each representation used for featurisation.
    For example, custom display names for a collection of three representations
    of which the second generates two features, would be represented as follows:
    [["a"], ["b0", "b1"], ["c"]]

    Display names provided by users should ideally already be in canonical form,
    but alternatives are accepted for convenience. This function parses them into
    canonical form:

    - a non-nested list: this can be used as a shorthand by users using a single
        representation (instead of having to create a single-molflux nested list)
    - a single molflux: a.k.a. a 'template' string. This can be used to provide
        a template that will be applied to all output names across all
        representations. If 'None', a default fallback template will be applied.

    At any level, template strings (or None) can be used to broadcast the
    template across all associated outputs. For example, the following canonical
    form display names will result in the same output display names:
        [["a"], [None, None], ["c"]]; [["a"], [None], ["c"]]
    The first example explicitly applies the default template to each output
    feature of the second representation, whereby the second example implicitly
    broadcasts the default template to all outputs (in case the number is not
    known in advance)

    Returns:
        The canonical form display names for downstream processing.
    """

    explanation_banner = """\n
    Display names should be provided as a nested list of display
    names, with one (list) entry for each representation used for featurisation.
    For example, custom display names for a collection of three representations
    of which the second generates two features, would be represented as follows:
    [["a"], ["b0", "b1"], ["c"]]
    """

    # avoids having to force 'representations' protocol to be Sized
    num_representations = len(list(representations))

    # 1) string template (or None): str_template -> [[str_template], ...]
    if display_names is None or isinstance(display_names, str):
        canonical_form = [[display_names] for _ in range(num_representations)]

    # 2) convenience converter for single representation notation: [x y ...] -> [[x y z]]
    elif isinstance(display_names, list) and all(
        x is None or isinstance(x, str) for x in display_names
    ):
        if num_representations != 1:
            raise ValueError(
                f"The provided display names syntax can only be used for single representations: {display_names!r} (you are using {num_representations})"
                + explanation_banner,
            )
        canonical_form = [display_names]  # type: ignore[list-item]

    # 3) inputs already in canonical form:
    elif isinstance(display_names, list) and all(
        isinstance(x, list) for x in display_names
    ):
        canonical_form = display_names  # type: ignore[assignment]

    # Reject unsupported free-form input syntax
    else:
        raise ValueError(
            f"The provided display names do not follow the accepted syntax: {display_names!r} (for {num_representations} representation(s))"
            + explanation_banner,
        )

    # Assert outputs are in canonical form (internal railguard)
    if (
        not all(isinstance(x, list) for x in canonical_form)
        or not len(canonical_form) == num_representations
    ):
        raise RuntimeError(
            f"Error parsing display names: {display_names!r} >> {canonical_form!r} (for {num_representations} representation(s))",
        )

    return canonical_form


def _consolidate_map_outputs(dataset: DatasetType) -> DatasetType:
    """Consolidates datasets returned by a `datasets.map` operation.

    One potential problem is that internally empty datasets are not processed
    by the `datasets.map` call. This can lead to inconsistencies in output
    feature names across empty and non-empty datasets.
    """
    if isinstance(dataset, datasets.Dataset):
        return dataset

    empty_splits = [k for k, v in dataset.num_rows.items() if v == 0]
    if not empty_splits:
        return dataset

    non_empty_splits = [k for k, v in dataset.num_rows.items() if v != 0]
    if not non_empty_splits:
        logger.warning(
            "Could not consolidate DatasetDict output features: all Datasets in the dict are empty",
        )
        return dataset

    # Replace all empty splits with empty datasets matching non-empty datasets
    reference = non_empty_splits[0]
    for k in empty_splits:
        empty_clone = datasets.Dataset(dataset[reference].data.slice(0, 0))
        dataset[k] = empty_clone

    return dataset


def featurise_dataset(
    dataset: DatasetType,
    column: str | list[str],
    representations: Representation | Representations,
    display_names: FreeformDisplayNames = None,
    **map_kwargs: Any,
) -> DatasetType:
    """Featurises a dataset column according to the given representations.

    Args:
        dataset: The dataset to featurise.
        column: One or multiple columns from the dataset to use as input to featurisers.
        representations: The representation or representations to featurise the columns with.
        display_names: A list of custom labels to assign to the newly
            featurised columns, or a single string template that will be
            used to dynamically generate labels.
        map_kwargs: Optional keyword arguments to be passed to the underlying
            dataset's .map() method during featurisation.

    Returns:
        The featurised dataset.
    """

    # make sure we're working with a list of columns
    columns = [column] if isinstance(column, str) else column

    # Make sure we can always iterate over a collection
    if not isinstance(representations, Representations):
        representations = [representations]

    # Make sure that all Datasets in the DatasetDict have the same features
    if isinstance(dataset, datasets.DatasetDict):
        all_features_match = len(set(map(tuple, dataset.column_names.values()))) == 1
        if not all_features_match:
            raise ValueError(
                f"Inconsistent input features across splits: got {dataset.column_names!r}",
            )

    canonical_display_names = _to_canonical_display_names(
        display_names,
        representations=representations,
    )

    # Apply batched featurisation instead of row-by-row
    if "batched" not in map_kwargs:
        map_kwargs["batched"] = True

    try:
        featurized_dataset = dataset.map(
            function=_featurise_batch,
            fn_kwargs={
                "columns": columns,
                "representations": representations,
                "display_names": canonical_display_names,
            },
            **map_kwargs,
        )
    except pa.ArrowInvalid as e:
        raise TypeError(
            "Apache Arrow serialisation error: one or more representations might be incompatible with Apache Arrow backends.",
        ) from e

    featurized_dataset = _consolidate_map_outputs(featurized_dataset)

    return featurized_dataset


def _featurise_batch(
    example: dict[str, Any],
    columns: list[str],
    representations: Representations,
    display_names: DisplayNames,
) -> dict[str, Any]:
    """Featurises a batch's column according to the given representations.

    The primary objective of batch mapping is to speed up processing. Often
    times, it is faster to work with batches of data instead of single examples.
    """

    for column in columns:
        if column not in example:
            raise KeyError(
                f"Feature {column!r} not in dataset: available features are {list(example.keys())!r}",
            )

    # get the relevant columns from the dataset
    samples = [example[column] for column in columns]

    for representation, representation_display_names in zip(
        representations,
        display_names,
        strict=False,
    ):
        # this could be a multi-output feature
        representation_results = representation.featurise(*samples)

        # handle template strings as display names for one-to-many representations
        if len(representation_display_names) == 1 and len(representation_results) != 1:
            display_name = representation_display_names[0]
            is_template = display_name is None or (
                "{" in display_name and "}" in display_name
            )
            if not is_template:
                raise ValueError(
                    f"Only template display names can be used as placeholders to be broadcasted: {display_name!r}",
                )
            representation_display_names = display_name  # type: ignore[assignment]

        # templates should be broadcasted to apply to all columns of the variably sized output
        for (feature_name, result), display_name in zip_broadcast(  # type: ignore[misc]
            representation_results.items(),  # type: ignore[arg-type]
            representation_display_names,  # type: ignore[arg-type]
        ):
            ctx = {"source_column": ":".join(columns), "feature_name": feature_name}  # type: ignore[has-type]
            featurised_column_name = _template(
                display_name,
                default_template=_DEFAULT_DISPLAY_NAMES_TEMPLATE,
                **ctx,
            )

            if featurised_column_name in example:
                warnings.warn(
                    f"An existing column is being overwritten on featurisation: {':'.join(columns)}::{feature_name} >> {featurised_column_name}",  # type: ignore[has-type]
                    stacklevel=1,
                )

            example[featurised_column_name] = result  # type: ignore[has-type]

    return example


def _template(target: str | None, default_template: str, **ctx: Any) -> str:
    """Templates a string with local context information.

    Args:
        target: The string to template. If None, the default template is used.
        default_template: The default template to use.
        ctx: Optional context that will be used to template the string.

    Returns:
        The templated string.
    """
    if target is None:
        target = default_template

    try:
        templated = target.format(**ctx)
    except KeyError as e:
        raise KeyError(
            f"Could not template the target string: {e} missng from context: {ctx!r}.",
        ) from e

    return templated



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.core.generic.exploded import Exploded

representation_name = "exploded"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, Exploded)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_representation_tag_matches_entrypoint_name(fixture_representation):
    """That the default representation tag matches the entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    representation = fixture_representation
    assert representation.tag == representation_name


def test_enumerate_fingerprints(fixture_representation):
    """That works as expected on fingerprint-like inputs."""
    representation = fixture_representation
    samples = [[1, 0, 0, 0], [1, 1, 0, 0], [1, 1, 1, 0], [1, 1, 1, 1]]
    results = representation.featurise(samples)

    # should return a column with all the first molfluxs, one with all the
    # second molfluxs, and so on...
    expected_results = [[1, 1, 1, 1], [0, 1, 1, 1], [0, 0, 1, 1], [0, 0, 0, 1]]
    for actual, expected in zip(results.values(), expected_results, strict=False):
        assert actual == expected



from typing import Any

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import datasets
from molflux.modelzoo.utils import (
    get_concatenated_array,
    raise_on_unrecognised_parameters,
)


def test_raise_on_unrecognised_parameters_raises_on_unrecognised_parameters():
    """That an error is raised if passing parameters not defined by the
    function."""

    def my_callable(a: str, b: int = 2, c: bool = False, **kwargs: Any) -> None: ...

    kwargs = {"b": 3, "not specified": "any"}
    with pytest.raises(ValueError, match=r"Unknown my_callable parameter\(s\)"):
        raise_on_unrecognised_parameters(my_callable, **kwargs)


def test_raise_on_unrecognised_parameters_does_not_raise_on_recognised_parameters():
    """That an error is not raised if passing parameters defined by the
    function."""

    def my_callable(a: str, b: int = 2, c: bool = False, **kwargs: Any) -> None: ...

    kwargs = {"a": "a", "b": 3}
    raise_on_unrecognised_parameters(my_callable, **kwargs)
    assert True


def test_get_concatenated_array_concatenates_nested_arrays():
    """That the function generates a single feature vector out of features with
    nested arrays.
    """
    dataset = datasets.Dataset.from_dict(
        {"a": [1, 2], "b": [[10, 11, 12], [20, 21, 22]]},
    )
    out = get_concatenated_array(dataset, features=["a", "b"])
    expected = np.array([[1, 10, 11, 12], [2, 20, 21, 22]])
    assert_array_equal(expected, out)


def test_get_concatenated_array_preserves_order():
    """That the function preserves the specified order in the output array"""
    dataset = datasets.Dataset.from_dict(
        {"a": [1, 2], "b": [[10, 11, 12], [20, 21, 22]]},
    )
    out = get_concatenated_array(dataset, features=["b", "a"])
    expected = np.array([[10, 11, 12, 1], [20, 21, 22, 2]])
    assert_array_equal(expected, out)


def test_get_concatenated_array_robust_to_empty_dataset():
    """That the function can handle empty input datasets."""
    dataset = datasets.Dataset.from_dict({"a": [], "b": []})
    out = get_concatenated_array(dataset, features=["a", "b"])
    expected = np.array([]).reshape(-1, 1)
    assert_array_equal(expected, out)



import functools
import json
import logging
import os
from typing import Any

from cloudpathlib import AnyPath
from pydantic.v1 import BaseModel, Field

from molflux.core.environment import pip_working_set
from molflux.core.io import save_dict_to_json
from molflux.core.typing import PathLike

logger = logging.getLogger(__name__)

_FEATURISATION_METADATA_FILENAME = "featurisation_metadata.json"


class RepresentationConfig(BaseModel):
    name: str
    config: dict[str, Any] = Field(default_factory=dict)
    presets: dict[str, Any] = Field(default_factory=dict)
    as_: str | list[str | None] | None = Field(
        default_factory=lambda: [None],
        alias="as",
    )


class ColumnFeaturisationConfig(BaseModel):
    column: str
    representations: list[RepresentationConfig]


class FeaturisationMetadataV1(BaseModel):
    version: int = Field(1, const=True)
    config: list[ColumnFeaturisationConfig] = Field(default_factory=list)
    runtime: dict[str, Any] = Field(default_factory=pip_working_set)


class ColumnFeaturisationConfigV2(BaseModel):
    columns: list[str]
    representations: list[RepresentationConfig]


class FeaturisationMetadataV2(BaseModel):
    version: int = Field(2, const=True)
    config: list[ColumnFeaturisationConfigV2] = Field(default_factory=list)
    runtime: dict[str, Any] = Field(default_factory=pip_working_set)


def _placeholder_featurisation_metadata() -> dict[str, Any]:
    """Returns an empty placeholder featurisation metadata payload"""
    return FeaturisationMetadataV2(version=2).dict(
        exclude_defaults=False,
        exclude_unset=False,
        exclude_none=False,
    )


def parse_featurisation_metadata(
    featurisation_metadata: dict[str, Any],
) -> tuple[dict[str, Any], int]:
    """Performs soft validation of featurisation metadata based on schema version."""

    if "version" not in featurisation_metadata:
        raise KeyError("Featurisation metadata doesn't specify a 'version'")

    version = featurisation_metadata["version"]
    if version == 1:
        metadata = FeaturisationMetadataV1(**featurisation_metadata).dict(by_alias=True)
    elif version == 2:
        metadata = FeaturisationMetadataV2(**featurisation_metadata).dict(by_alias=True)
    else:
        raise NotImplementedError(
            f"Unsupported featurisation metadata version: {version!r}",
        )

    return metadata, version


def save_featurisation_metadata(
    featurisation_metadata: dict[str, Any] | None,
    path: PathLike,
) -> str:
    """Saves featurisation metadata to disk.

    Args:
        featurisation_metadata: The featurisation metadata payload to be saved.
            If None or empty, an empty placeholder metadata schema is saved.
        path: The path under which to save the featurisation metadata.
    """

    if not featurisation_metadata:
        featurisation_metadata = _placeholder_featurisation_metadata()

    # Validate payload and fill-in
    featurisation_metadata, _ = parse_featurisation_metadata(featurisation_metadata)

    output_file = f"{path}/{_FEATURISATION_METADATA_FILENAME}"
    return save_dict_to_json(featurisation_metadata, path=output_file)


def fetch_model_featurisation_metadata(model_path: str) -> dict[str, Any]:
    """Retrieves the featurisation metadata associated with a given model.

    Args:
        model_path: The path under which the model of interest is saved.

    Returns:
        The featurisation metadata that was used to generate the model's input
        features.
    """

    # Expected to be saved alongside the model artefact
    expected_metadata_path = os.path.join(model_path, _FEATURISATION_METADATA_FILENAME)
    return load_featurisation_metadata(expected_metadata_path)


def load_featurisation_metadata(path: str) -> dict[str, Any]:
    """Loads json featurisation metadata from a given path."""
    return _cached_load_featurisation_metadata(path).copy()


@functools.lru_cache
def _cached_load_featurisation_metadata(path: str) -> dict[str, Any]:
    """Loads json featurisation metadata from a given path, caching the result"""

    try:
        with AnyPath(path).open("r") as f:  # type: ignore[attr-defined]
            featurisation_metadata = json.load(f)
        logger.info(f"Fetched featurisation metadata: {path}")
    except Exception as error:
        logger.exception(error)
        raise FileNotFoundError(
            f"Error fetching featurisation metadata: could not fetch {path}",
        ) from error

    return featurisation_metadata  # type:ignore[no-any-return]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "molflux"
author = "Exscientia"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx_toolbox.confval",
    "sphinx_togglebutton",
    "myst_nb",
    "sphinx_design",
    "sphinx_sitemap",
    "sphinx_inline_tabs",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_baseurl = "https://exscientia.github.io/molflux/"
html_favicon = "_static/MolFlux_logo_final_MolFlux logo mark.svg"

html_title = "MolFlux"

html_theme_options = {
    "logo": {
        "image_dark": "_static/MolFlux_logo_final_MolFlux Logo mark with name 2 colours.svg",
        "image_light": "_static/MolFlux_logo_final_MolFlux Logo mark with name grey.svg",
    },
    "repository_url": "https://github.com/Exscientia/molflux",
    "use_repository_button": True,
    "use_sidenotes": True,
    "show_nav_level": 0,
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/molflux/",
            "icon": "https://img.shields.io/pypi/v/molflux",
            "type": "url",
        },
    ],
}



from molflux.metrics.bases import UncertaintyMetric
from molflux.metrics.metric import Metric, Metrics


def supports_prediction_intervals(metric: Metric) -> bool:
    """Return True if the given metric supports uncertainty predictions."""
    return isinstance(metric, UncertaintyMetric)


def all_support_prediction_intervals(metrics: Metrics) -> bool:
    """Return True if all the given metric supports uncertainty predictions."""
    return all(supports_prediction_intervals(metric) for metric in metrics)



from typing import Any

try:
    from openeye import oechem
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("openeye", e) from None
from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.openeye._utils import smiles_from_oemol, to_oemol
from molflux.features.representations.openeye.canonical._utils import standardise_oemol
from molflux.features.typing import MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
A canonical SMILES representation.
"""


class CanonicalSmiles(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        strip_salts: bool = False,
        set_neutral_ph: bool = False,
        reasonable_protomer: bool = False,
        reasonable_tautomer: bool = False,
        explicit_h: bool = False,
        remove_formal_charges: bool = False,
        assign_hyb: bool = True,
        rekekulise: bool = False,
        remove_stereo: bool = False,
        clear_non_chiral_stereo: bool = False,
        remove_non_standard_stereo: bool = False,
        sd_title_tag: str = "",
        clear_sd_data: bool = False,
        tautomer_options: dict[str, Any] | None = None,
        tautomer_timeouts: list[float] | None = None,
        **kwargs: Any,
    ) -> dict[str, list[str]]:
        """Featurises the input molecules as canonical SMILES strings.

        Args:
            samples: The sample molecules to featurise.
            strip_salts: Whether to remove everything but the largest component
                from the molecules, in case they contain disconnected
                components such as salts. Defaults to `False`.
            set_neutral_ph: Whether to set a neutral pH model to the molecules.
                Defaults to `False`.
            reasonable_protomer: Whether to attempt to produce a single protomer
                that will be a suitable representation of the molecule in a
                biological system. This state is defined as an aqueous
                environment with pH~7.4 and a tautomer from among the
                predominate tautomers that is favored by medicinal chemists.
                Defaults to `False`.
            reasonable_tautomer: Whether to generate reasonable tautomers and
                select the top-ranked. Defaults to `False`.
            explicit_h: Whether to convert the implicit hydrogens on the atoms
                of a molecule to explicit hydrogen atoms. Defaults to `False`.
            remove_formal_charges: Whether to remove formal charges from the
                input molecules. Defaults to `False`.
            assign_hyb: Whether to assign hybridization to all atoms in the
                given molecules. Defaults to `True`.
            rekekulise: Whether to reassign the single-double bond distribution
                in aromatic rings. Defaults to `False`.
            remove_stereo: Whether to remove stereochemistry information from
                the input molecule. Defaults to `False`.
            clear_non_chiral_stereo: Whether to remove false atom stereo
                information from non-chiral atoms. Defaults to `False`.
            remove_non_standard_stereo: OpenEye's definition of what is chiral
                includes tertiary nitrogens. This isn't really from a synthetic
                chemists standpoint - so this option removes chiral centres from
                these if set to `True`. Defaults to `False`.
            sd_title_tag: The SD tag to use as the molecules' title. Defaults to
                the empty string.
            clear_sd_data: Whether to clear all SD data from the molecules.
                Defaults to `False`.
            tautomer_options: Dictionary of settings for the tautomerisation stage. Keys and value types allowed are:
                - max_generated: int
                - rank: bool
                - warts: bool
                - level: int
                - carbon_hybrid: bool
                - save_stereo: bool
                - max_zone_size: int
                - max_tautomeric_atoms: int
                - max_time: int
                - kekule: bool
                - clear_coords: bool
                - hybrid_level: int
                - max_to_return: int
                - racemic_type: int
                NOTE: This parameter has an effect only if `reasonable_tautomer` is set to True.
            tautomer_timeouts: List of timeouts to be passed to the tautomerisation stage.
                For each molecule, each timeout is tried at the tautomerisation step until the first reasonable
                tautomer is generated. Once one is found, the other timeouts are skipped.
                NOTE: This parameter has an effect only if `reasonable_tautomer` is set to True.

        Returns:
            inputs featurised as canonical SMILES strings.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation("canonical_smiles")
            >>> samples = ["CCCC"]
            >>> representation.featurise(samples)
            {'canonical_smiles': ['CCCC']}
        """

        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]

        canonical_smiles = []
        flavor = oechem.OEGetDefaultOFlavor(oechem.OEFormat_SMI)
        if explicit_h:
            flavor |= oechem.OEOFlavor_SMI_Hydrogens
        if rekekulise:
            flavor |= oechem.OEOFlavor_SMI_Kekule
        for sample in samples:
            with featurisation_error_harness(sample):
                mol = to_oemol(sample)
                canonical_mol = standardise_oemol(
                    mol,
                    strip_salts=strip_salts,
                    set_neutral_ph=set_neutral_ph,
                    reasonable_protomer=reasonable_protomer,
                    reasonable_tautomer=reasonable_tautomer,
                    explicit_h=explicit_h,
                    remove_formal_charges=remove_formal_charges,
                    perceive_chiral=True,
                    assign_hyb=assign_hyb,
                    rekekulize=rekekulise,
                    remove_stereo=remove_stereo,
                    clear_non_chiral_stereo=clear_non_chiral_stereo,
                    remove_non_standard_stereo=remove_non_standard_stereo,
                    sd_title_tag=sd_title_tag,
                    clear_sd_data=clear_sd_data,
                    gen_2d_coords=False,
                    tautomer_options=tautomer_options,
                    tautomer_timeouts=tautomer_timeouts,
                )
                smiles = smiles_from_oemol(canonical_mol, flavor=flavor)
                canonical_smiles.append(smiles)

        return {self.tag: canonical_smiles}



from typing import Any

import numpy as np
import pandas as pd

from molflux.splits.typing import ArrayLike
from molflux.splits.utils import partition


def validate_ordered_strategy_arguments(
    stratified_bool: bool,
    y: ArrayLike | None,
    groups: ArrayLike | None,
    train_fraction: float | list[float],
    gap_train_validation_fraction: float | list[float] | None,
    validation_fraction: float | list[float],
    gap_validation_test_fraction: float | list[float] | None,
    test_fraction: float | list[float],
    gap_test_end_fraction: float | list[float] | None,
    stratify_targets_to_ignore: list | None,
    min_test_size: int | list[int] | None,
) -> dict[str, Any]:
    """
    method to validate that all input data is correct.
        checks that groups exists, have > 2 groups, have same type, are comparable
        checks that train, validation, test, gaps, fractions are same type, same length,
            and fills haps with 0.0 if not defined
        checks that fractions add up to 1
        checks that targets exists (if specified) and finds targets to ignore
    """

    # assert that groups is defined
    if groups is None:
        raise ValueError("Need to define 'groups' for ordered splits.")

    # assert that there are more than 2 groups
    assert len(set(groups)) > 2, RuntimeError(
        "Not enough groups for ordered split. Must be > 2.",
    )

    groups = [x if not pd.isnull(x) else None for x in groups]

    # assert that all groups are of same type (or None)
    type_check = list({type(x) for x in groups})
    if type(None) in type_check:
        type_check.remove(type(None))
    assert len(type_check) == 1, TypeError(
        "Groups are not all of the same type (or None, nan, NaT)",
    )

    # assert that groups support comparison
    assert all(
        ((x is None) or (hasattr(x, "__ge__") and hasattr(x, "__le__"))) for x in groups
    ), TypeError("Group molfluxs cannot be compared.")

    assert (
        isinstance(train_fraction, list)
        and isinstance(validation_fraction, list)
        and isinstance(test_fraction, list)
    ) or (
        isinstance(train_fraction, (int, float))
        and isinstance(validation_fraction, (int, float))
        and isinstance(test_fraction, (int, float))
    ), TypeError(
        f"Type of train, validation, test are not the same {[type(train_fraction), type(validation_fraction), type(test_fraction)]}",
    )

    if isinstance(train_fraction, (int, float)):
        train_fraction_list: list = [train_fraction]
    else:
        train_fraction_list = train_fraction
    if isinstance(validation_fraction, (int, float)):
        validation_fraction_list: list = [validation_fraction]
    else:
        validation_fraction_list = validation_fraction
    if isinstance(test_fraction, (int, float)):
        test_fraction_list: list = [test_fraction]
    else:
        test_fraction_list = test_fraction

    if (
        len(
            {
                len(train_fraction_list),
                len(validation_fraction_list),
                len(test_fraction_list),
            },
        )
        != 1
    ):
        raise TypeError(
            f"Lengths of train, validation, test are not the same "
            f"{[len(train_fraction_list), len(validation_fraction_list), len(test_fraction_list)]}",
        )

    if gap_train_validation_fraction is None:
        gap_train_validation_fraction_list: list = [
            0.0 for _ in range(len(train_fraction_list))
        ]
    else:
        if isinstance(gap_train_validation_fraction, float):
            gap_train_validation_fraction_list = [gap_train_validation_fraction]
        else:
            assert isinstance(gap_train_validation_fraction, list)
            gap_train_validation_fraction_list = gap_train_validation_fraction

    if gap_validation_test_fraction is None:
        gap_validation_test_fraction_list: list = [
            0.0 for _ in range(len(train_fraction_list))
        ]
    else:
        if isinstance(gap_validation_test_fraction, float):
            gap_validation_test_fraction_list = [gap_validation_test_fraction]
        else:
            assert isinstance(gap_validation_test_fraction, list)
            gap_validation_test_fraction_list = gap_validation_test_fraction

    if gap_test_end_fraction is None:
        gap_test_end_fraction_list: list = [
            0.0 for _ in range(len(train_fraction_list))
        ]
    else:
        if isinstance(gap_test_end_fraction, float):
            gap_test_end_fraction_list = [gap_test_end_fraction]
        else:
            assert isinstance(gap_test_end_fraction, list)
            gap_test_end_fraction_list = gap_test_end_fraction

    if min_test_size is None:
        min_test_size_list: list = [0 for _ in range(len(test_fraction_list))]
    else:
        if isinstance(min_test_size, int):
            min_test_size_list = [min_test_size]
        else:
            assert isinstance(min_test_size, list)
            min_test_size_list = min_test_size

    for min_test, test_fraction in zip(
        min_test_size_list,
        test_fraction_list,
        strict=False,
    ):
        if test_fraction == 0.0:
            assert min_test == 0, RuntimeError(
                f"Cannot have non-zero min test size {min_test} when test fraction is {test_fraction}.",
            )

    # assert ratios are equal to 1
    for (
        tra_fr,
        g_tr_val_fr,
        val_fr,
        g_val_tes_fr,
        test_fr,
        g_tes_end_fr,
    ) in zip(
        train_fraction_list,
        gap_train_validation_fraction_list,
        validation_fraction_list,
        gap_validation_test_fraction_list,
        test_fraction_list,
        gap_test_end_fraction_list,
        strict=False,
    ):
        np.testing.assert_almost_equal(
            sum([tra_fr, g_tr_val_fr, val_fr, g_val_tes_fr, test_fr, g_tes_end_fr]),
            1.0,
        )

    if stratified_bool:
        # assert that targets is defined
        if y is None:
            raise ValueError("Need to define 'targets' for stratified ordered splits.")

        y = [x if not pd.isnull(x) else None for x in y]

        y_expanded = []
        groups_expanded = []
        indices_expanded = []

        # check same length
        if len(list(y)) != len(list(groups)):
            raise RuntimeError("Length of groups is not equal to length of targets")

        for idx, (yy, group) in enumerate(zip(y, groups, strict=False)):
            if isinstance(yy, str):
                yy_split = [x.strip() for x in yy.split("|")]
            else:
                yy_split = [yy]

            if len(yy_split) > 1:
                y_expanded += yy_split
                groups_expanded += [group] * len(yy_split)
                indices_expanded += [idx] * len(yy_split)
            else:
                y_expanded.append(yy_split[0])
                groups_expanded.append(group)
                indices_expanded.append(idx)

        y = y_expanded
        groups = groups_expanded
        indices = indices_expanded

        # get set of targets
        set_of_targets: set = set(y)

        if None in set_of_targets:
            set_of_targets.remove(None)

        # if need to ignore some targets
        if stratify_targets_to_ignore is not None:
            assert isinstance(stratify_targets_to_ignore, list), TypeError(
                f"'stratify_targets_to_ignore' {stratify_targets_to_ignore} must be a list of targets to stratify by.",
            )

            set_of_targets = set_of_targets - set(stratify_targets_to_ignore)

        assert len(set_of_targets) > 1, RuntimeError(
            "Need more than 1 targets to stratify with.",
        )
        groups_array: np.ndarray = np.array(groups)
        targets_array: np.ndarray = np.array(y)
        index_array: np.ndarray = np.array(indices)

    else:
        set_of_targets = set()
        # define groups array, targets array, index array
        groups_array = np.array(groups)
        targets_array = np.array(y)
        index_array = np.arange(len(list(groups)))

    validated_dict = {
        "train_fraction_list": train_fraction_list,
        "gap_train_validation_fraction_list": gap_train_validation_fraction_list,
        "validation_fraction_list": validation_fraction_list,
        "gap_validation_test_fraction_list": gap_validation_test_fraction_list,
        "test_fraction_list": test_fraction_list,
        "gap_test_end_fraction_list": gap_test_end_fraction_list,
        "set_of_targets": set_of_targets,
        "groups_array": groups_array,
        "targets_array": targets_array,
        "index_array": index_array,
        "min_test_size_list": min_test_size_list,
    }

    return validated_dict


def make_sorted_groups_indices(
    groups_array: np.ndarray,
    index_array: np.ndarray,
    targets_array: np.ndarray | None,
    target: str | None,
) -> dict:
    """
    method to find sorted groups and indices with and without None groups
    """
    if (target is not None) and (targets_array is not None):
        # bool mask of this target and groups Not equal to None
        mask_without_nones = (targets_array == target) * (
            groups_array != None  # noqa: E711
        )

        # bool mask of this target and groups equal to None
        mask_of_nones = (targets_array == target) * (groups_array == None)  # noqa: E711

    else:
        # bool mask of groups Not equal to None
        mask_without_nones = groups_array != None  # noqa: E711

        # bool mask of groups equal to None
        mask_of_nones = groups_array == None  # noqa: E711

    # find target indices and indices of Nones
    indices_without_nones = index_array[mask_without_nones]
    indices_of_nones = index_array[mask_of_nones]

    # find target groups
    groups_without_nones = groups_array[mask_without_nones]
    if len(groups_without_nones) > 0:
        groups_without_nones = groups_without_nones.astype(
            type(groups_without_nones[0]),
        )
    groups_of_nones = groups_array[mask_of_nones]

    # find sorted indices based on groups
    sorted_args_without_nones = np.argsort(groups_without_nones, kind="stable")
    sorted_groups_without_nones = groups_without_nones[sorted_args_without_nones]
    sorted_indices_without_nones = indices_without_nones[sorted_args_without_nones]

    # final sorted groups and indices with nones at the beginning
    sorted_groups_with_nones = np.concatenate(
        [groups_of_nones, sorted_groups_without_nones],
    )
    sorted_indices_with_nones = np.concatenate(
        [indices_of_nones, sorted_indices_without_nones],
    )

    sorted_groups_indices_dict: dict = {
        "sorted_indices_without_nones": sorted_indices_without_nones,
        "sorted_indices_with_nones": sorted_indices_with_nones,
        "sorted_groups_without_nones": sorted_groups_without_nones,
        "sorted_groups_with_nones": sorted_groups_with_nones,
    }

    return sorted_groups_indices_dict


def find_begin_end_indices(
    sorted_indices_without_nones: np.ndarray,
    sorted_indices_with_nones: np.ndarray,
    split_train_fraction: float,
    split_gap_train_val_fraction: float,
    split_validation_fraction: float,
    split_gap_val_test_fraction: float,
    split_test_fraction: float,
    split_gap_test_end_fraction: float,
    undefined_groups_in_train: bool,
) -> dict:
    """method to find begin, end, indices of different sets"""

    if undefined_groups_in_train:
        # partitioning is performed on the dataset with None-groups ignored...
        indices = sorted_indices_without_nones
    else:
        indices = sorted_indices_with_nones

    train_end, validation_begin, validation_end, test_begin, test_end = partition(
        indices,
        split_train_fraction,
        split_gap_train_val_fraction,
        split_validation_fraction,
        split_gap_val_test_fraction,
        split_test_fraction,
    )

    cutoffs: dict[str, int] = {
        "train_end": train_end,
        "validation_begin": validation_begin,
        "validation_end": validation_end,
        "test_begin": test_begin,
        "test_end": test_end,
    }

    # ... and then we inject the Nones in the train section
    # (cascading the offset introduced by the extra items)
    if undefined_groups_in_train:
        n_nones = len(sorted_indices_with_nones) - len(sorted_indices_without_nones)
        cutoffs = {k: v + n_nones for k, v in cutoffs.items()}

    return cutoffs


def force_min_test_size(
    min_test_size: int,
    train_end: int,
    validation_begin: int,
    validation_end: int,
    test_begin: int,
    test_end: int,
) -> dict:
    """
    method to force test set to have min size. Shifts test begin and modifies
    validation and train such that train is used up first, then validation
    """
    if min_test_size > test_end - test_begin:
        old_test_begin = test_begin

        # If min_test_size is bigger that num_indices, everything goes in test set
        if min_test_size > test_end:
            test_begin = 0
        else:
            test_begin = test_end - min_test_size

        # how much to shift indices of other sets
        shift_indices = old_test_begin - test_begin

        # shift sets and set to 0 if outside of range
        validation_end = max(validation_end - shift_indices, 0)
        validation_begin = max(validation_begin - shift_indices, 0)
        train_end = max(train_end - shift_indices, 0)

    bgn_end_dict: dict = {
        "train_end": train_end,
        "validation_begin": validation_begin,
        "validation_end": validation_end,
        "test_begin": test_begin,
        "test_end": test_end,
    }

    return bgn_end_dict


def force_same_set_same_group(
    sorted_groups_with_nones: np.ndarray,
    terminal_index: int,
) -> int:
    """
    method to force the beginning/end of sets to correspond to whole groups.
    """

    if terminal_index == 0:
        return terminal_index
    elif terminal_index == len(sorted_groups_with_nones):
        return terminal_index
    else:
        # what is the group
        terminal_group = sorted_groups_with_nones[terminal_index - 1]

        # find group indices
        terminal_group_indices = np.where(sorted_groups_with_nones == terminal_group)[0]

        # pick nearest
        terminal_index = min(
            [
                terminal_group_indices[0],
                terminal_group_indices[-1] + 1,
            ],
            key=lambda x: int(abs(x - terminal_index)),
        )

        return terminal_index



import pytest

from molflux.features.catalogue import list_representations
from molflux.features.errors import ExtrasDependencyImportError
from molflux.features.load import load_representation

_PLUGIN_REPRESENTATIONS = {
    group: names for group, names in list_representations().items() if group != "core"
}


@pytest.mark.parametrize(
    "extra,representation_name",
    [
        (group, name)
        for group in _PLUGIN_REPRESENTATIONS
        for name in _PLUGIN_REPRESENTATIONS[group]
    ],
)
def test_missing_extra_dependencies_raise_on_load(extra, representation_name):
    """That if dependencies are missing for representations offered as an extra,
    an error is raised at load time.

    The error should hint at how to install the missing dependencies.
    """

    with pytest.raises(
        ExtrasDependencyImportError,
        match=rf"pip install \'molflux\[{extra}\]\'",
    ):
        load_representation(representation_name)



from typing import Any

try:
    from rdkit.Chem.AllChem import GetMorganFingerprintAsBitVect

    from molflux.features.representations.rdkit._utils import rdkit_mol_from_smiles
except ImportError as e:
    from molflux.features.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("rdkit", e) from None

from molflux.features.bases import RepresentationBase
from molflux.features.info import RepresentationInfo
from molflux.features.representations.rdkit._utils import to_smiles
from molflux.features.typing import Fingerprint, MolArray
from molflux.features.utils import assert_n_positional_args, featurisation_error_harness

_DESCRIPTION = """
Morgan fingerprint.

These fingerprints are similar to the well-known ECFP or FCFP fingerprints,
depending on which invariants are used. These are implemented based on the
original paper. The algorithm follows the description in the paper as
closely as possible with the exception of the chemical feature definitions used
for the “Feature Morgan” fingerprint - the RDKit implementation uses the
feature types Donor, Acceptor, Aromatic, Halogen, Basic, and Acidic with
definitions adapted from those in [1]_. It is possible to provide your
own atom types. The fingerprints are available as either explicit or sparse
count vectors or explicit bit vectors.

The algorithm used is described in the paper
Rogers, D. & Hahn, M. Extended-Connectivity Fingerprints. JCIM 50:742-54 (2010)
https://doi.org/10.1021/ci100050t

The original implementation was done using this paper:
D. Rogers, R.D. Brown, M. Hahn J. Biomol. Screen. 10:682-6 (2005)
and an unpublished technical report:
http://www.ics.uci.edu/~welling/teaching/ICS274Bspring06/David%20Rogers%20-%20ECFP%20Manuscript.doc

[1]_ https://doi.org/10.1002/(SICI)1097-0290(199824)61:1%3C47::AID-BIT9%3E3.0.CO;2-Z
"""


class Morgan(RepresentationBase):
    def _info(self) -> RepresentationInfo:
        return RepresentationInfo(
            description=_DESCRIPTION,
        )

    def _featurise(
        self,
        *columns: MolArray,
        radius: int = 3,
        n_bits: int = 2048,
        invariants: list[int] | None = None,
        from_atoms: list[int] | None = None,
        use_chirality: bool = False,
        use_bond_types: bool = True,
        use_features: bool = False,
        bit_info: dict | None = None,
        include_redundant_environments: bool = False,
        **kwargs: Any,
    ) -> dict[str, list[Fingerprint]]:
        """Generates Morgan fingerprints for each input molecule.

        Args:
            samples: The molecules to be fingerprinted.
            radius: The number of iterations to grow the fingerprint.
            n_bits: The size of the fingerprint. Defaults to `2048`.
            invariants: The set of atom invariants to be used. Defaults to
                `None`, which corresponds to ECFP-type invariants.
            from_atoms: If provided, only the atoms in the vector will be used
                as centers in the fingerprint. Defaults to `None`.
            use_chirality: If set, additional information will be added to the
                fingerprint when chiral atoms are discovered, generating
                different fingerprints. Defaults to `False`.
            use_bond_types: If set, bond types will be included as part of the
                hash for calculating bits. Defaults to `True`.
            use_features: Defaults to `False`.
            bit_info: Defaults to `None`.
            include_redundant_environments: If not None, the check for redundant
                atom environments will not be done. Defaults to `False`.

        Returns:
            MACCS fingerprints, as lists of bits.

        Examples:
            >>> from molflux.features import load_representation
            >>> representation = load_representation('morgan')
            >>> samples = ['c1ccccc1']
            >>> representation.featurise(samples, n_bits=16)
            {'morgan': [[1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]}
        """
        assert_n_positional_args(*columns, expected_size=1)
        samples = columns[0]
        morgan_fp_list: list[list] = []
        for sample in samples:
            with featurisation_error_harness(sample):
                smile = to_smiles(sample)
                mol = rdkit_mol_from_smiles(smile)
                rd_fp = GetMorganFingerprintAsBitVect(
                    mol,
                    radius=radius,
                    nBits=n_bits,
                    invariants=invariants or [],
                    fromAtoms=from_atoms or [],
                    useChirality=use_chirality,
                    useBondTypes=use_bond_types,
                    useFeatures=use_features,
                    bitInfo=bit_info or {},
                    includeRedundantEnvironments=include_redundant_environments,
                )

                if rd_fp.GetNumOnBits() == 0:
                    morgan_fp_list.append([0] * rd_fp.GetNumBits())
                else:
                    morgan_fp_list.append(rd_fp.ToList())

        return {self.tag: morgan_fp_list}



from molflux.modelzoo.catalogue import fill_catalogue, list_models, register_model
from molflux.modelzoo.load import (
    load_from_dict,
    load_from_dicts,
    load_from_yaml,
    load_model,
)
from molflux.modelzoo.model import ClassificationMixin, ModelBase
from molflux.modelzoo.protocols import (
    Model,
    Models,
    supports_classification,
    supports_covariance,
    supports_prediction_interval,
    supports_sampling,
    supports_std,
    supports_uncertainty_calibration,
)
from molflux.modelzoo.store.io import load_from_store, save_to_store

# Register all plugins at package import time (to fill catalogue)
fill_catalogue()



"""Computes an ROC curve over SMILES strings, i.e. "is the correct synthesis product in top-k predictions"."""

import logging
from typing import Any

import evaluate
import numpy as np

import datasets
from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_CITATION = """\
@InProceedings{exscientia:accuracy_roc,
title = {Accuracy ROC},
authors={Exscientia},
year={2022}
}
"""

_DESCRIPTION = """\
Computes an accuracy ROC curve over SMILES strings, i.e. "is the correct synthesis product in top-k predictions".
"""


# TODO: Add description of the arguments of the metric here
_KWARGS_DESCRIPTION = """
Calculates what proportion of SMILES-encoded predictions hit the correct answer with the top k tries.

Args:
    predictions: list of lists of predictions to score. Each prediction should be a SMILES string.
    references: the correct answer for each set of predictions. Each
        reference should be a SMILES string.
    average: If `True`, an molflux-wise average across results is performed.

Returns:
    accuracy_roc: list whose k component is the answer to "what proportion
        of examples placed the correct answer within the first k predictions?"
        If 'average=True', "on average, what proportion of examples placed the correct answer within the first k predictions?"

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("top_k_accuracy_roc")
    >>> predictions = [['A', 'B', 'C', 'D', 'E']]
    >>> references = ['B']
    >>> metric.compute(predictions=predictions, references=references)
    {'top_k_accuracy_roc': [[0.0, 1.0, 1.0, 1.0, 1.0]]}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class TopKAccuracyRoc(HFMetric):
    """Computes the accuracy ROC curve.

    Useful for synthesis prediction."""

    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            # This is the description that will appear on the metrics page.
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            # This defines the format of each prediction and reference
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("string")),
                    "references": datasets.Value("string"),
                },
            ),
        )

    def _pre_process_inputs(
        self,
        predictions: ArrayLike | None = None,
        references: ArrayLike | None = None,
    ) -> tuple[ArrayLike | None, ...]:
        """Converts None inputs to strings to allow using None as invalid token.

        Otherwise, None molfluxs would raise as they do not count as string features.
        """
        if predictions is not None:
            predictions = [
                [p if p is not None else "None" for p in ps] for ps in predictions
            ]

        if references is not None:
            references = [r if r is not None else "None" for r in references]

        return predictions, references

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        average: bool = False,
        **kwargs: Any,
    ) -> MetricResult:
        """Returns the accuracy ROC curve"""

        max_k = len(predictions[0])

        def get_pred_idx(preds: list[str], correct: str) -> int:
            # Return max_k if none of the predictions match
            return next((i for i, pred in enumerate(preds) if pred == correct), max_k)

        idxs = [
            get_pred_idx(preds, product)
            for preds, product in zip(predictions, references, strict=False)
        ]

        # first for each idx in idxs define a vector that is 0 until idx (exclusive) and 1 after
        # note that max_k is larger than any i in range(max_k) so float(i >= max_k) = 0.
        present_at_idx = np.array(
            [[float(j >= idx) for j in range(max_k)] for idx in idxs],
        )

        if average:
            present_at_idx = present_at_idx.mean(axis=0, keepdims=True)

        accuracy_roc: list[float] = present_at_idx.tolist()

        return {self.tag: accuracy_roc}



"""
Tests ensuring desired API objects are part of the top-level namespace.
"""

import pytest

import molflux.core
import molflux.core.tracking


def test_exports_version():
    """That the package exposes the __version__ variable."""
    assert hasattr(molflux, "__version__")


@pytest.mark.parametrize(
    "callable_name",
    [
        "featurise_dataset",
        "replay_dataset_featurisation",
        "fetch_model_featurisation_metadata",
        "load_featurisation_metadata",
        "get_inputs",
        "get_references",
        "inference",
        "load_model",
        "predict",
        "save_model",
        "compute_scores",
        "invert_scores_hierarchy",
        "merge_scores",
        "score_model",
    ],
)
def test_exports_callable(callable_name):
    """That the package exposes the given callable."""
    assert hasattr(molflux.core, callable_name)
    assert callable(getattr(molflux.core, callable_name))


def test_exports_tracking_api():
    """That the package exposes the .tracking module."""
    assert hasattr(molflux.core, "tracking")


@pytest.mark.parametrize(
    "tracking_api_callable_name",
    [
        "log_dataset",
        "log_dataset_dict",
        "log_featurised_dataset",
        "log_featurisation_metadata",
        "log_fold",
        "log_inputs",
        "log_model_params",
        "log_params",
        "log_predictions",
        "log_pipeline_config",
        "log_references",
        "log_scores",
        "log_splitting_strategy",
    ],
)
def test_exports_tracking_api_callable(tracking_api_callable_name):
    """That the package .tracking API exposes the given callable."""
    assert hasattr(molflux.core.tracking, tracking_api_callable_name)
    assert callable(getattr(molflux.core.tracking, tracking_api_callable_name))



import os
import pathlib

import _pytest.config
import boto3
import fsspec
import pytest
from moto.server import ThreadedMotoServer

PytestConfig = type[_pytest.config.Config]


def path_to_file(pytestconfig: PytestConfig, *other: str) -> pathlib.Path:
    """Get an absolute path to the given target file."""
    return pytestconfig.rootpath.joinpath(*other)  # type: ignore


@pytest.fixture(scope="module")
def fixture_path_to_assets(pytestconfig: PytestConfig) -> pathlib.Path:
    """Get an absolute path to the assets directory."""
    return path_to_file(pytestconfig, "tests", "datasets", "assets")


@pytest.fixture(scope="module")
def fixture_path_to_this_test(request) -> pathlib.Path:  # type:ignore[no-untyped-def]
    """Get an absolute path to the module invoking this fixture.

    'request' is a magic pytest fixture which contains all information about
    the current test and fixture construction.
    """
    return pathlib.Path(os.path.dirname(request.module.__file__))


@pytest.fixture(scope="module")
def fixture_path_to_output(pytestconfig: PytestConfig) -> pathlib.Path:
    """Get an absolute path to the output directory"""
    return path_to_file(pytestconfig, "tests", "output")


@pytest.fixture(scope="module")
def _fixture_mock_moto_server():
    """A moto server to mock S3 interactions.

    We cannot use the usual moto decorators because pyarrow's S3 FileSystem
    is not based on boto3 at all.

    References:
        https://issues.apache.org/jira/browse/ARROW-16437
    """

    server = ThreadedMotoServer()
    try:
        server.start()
        yield
    finally:
        server.stop()


@pytest.fixture()
def _fixture_mock_aws_environment(monkeypatch):
    """A mock AWS credentials environment."""
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "testing")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "testing")
    monkeypatch.setenv("AWS_SECURITY_TOKEN", "testing")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "testing")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")
    monkeypatch.setenv("AWS_REGION", "us-east-1")


@pytest.fixture()
def fixture_mock_s3_client(_fixture_mock_aws_environment, _fixture_mock_moto_server):
    """An S3 client to the mocked moto AWS server."""
    client = boto3.client("s3", endpoint_url="http://localhost:5000")
    return client


@pytest.fixture()
def fixture_mock_s3_filesystem(_fixture_mock_aws_environment):
    """An S3 FileSystem connected to the mocked AWS server"""
    fs = fsspec.filesystem("s3", endpoint_override="http://localhost:5000")
    return fs



import pytest
from rdkit.Chem import Mol

from molflux.features.representations.rdkit._utils import (
    rdkit_mol_from_smiles,
    to_rdkit_mol,
    to_smiles,
)

representation_name = "avalon"


def test_rdkit_mol_from_smiles_valid_smiles():
    """Test that rdkit_mol_from_smiles works on valid data"""
    valid_smiles = [
        "c1ccccc1",
        "CCC",
        "CCCC1=NN(C2=C1NC(=NC2=O)C3=C(C=CC(=C3)S(=O)(=O)N4CCN(CC4)C)OCC)C",
    ]

    mols = [rdkit_mol_from_smiles(smiles) for smiles in valid_smiles]

    for mol in mols:
        assert mol is not None
        assert isinstance(mol, Mol)


def test_bytes_to_rdkit_mol():
    """That can convert a bytes string to OEMol object"""
    # Acetic Acid
    bytestring = b"\xef\xbe\xad\xde\x00\x00\x00\x00\r\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x80\x01\x06\x00`\x00\x00\x00\x01\x03\x06\x00(\x00\x00\x00\x03\x04\x08\x00(\x00\x00\x00\x03\x02\x08\x00h\x00\x00\x00\x03\x01\x01\x0b\x00\x01\x00\x01\x02(\x02\x01\x03 \x14\x00\x00\x00\x00\x17\x04\x00\x00\x00\x00\x00\x00\x00\x16"

    out = to_rdkit_mol(bytestring)
    assert isinstance(out, Mol)

    out_as_smiles = to_smiles(out)
    assert out_as_smiles == "CC(=O)O"


def test_rdkit_bytes_to_smiles():
    """That can convert a bytes string to SMILES"""
    # Acetic Acid
    bytestring = b"\xef\xbe\xad\xde\x00\x00\x00\x00\r\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x80\x01\x06\x00`\x00\x00\x00\x01\x03\x06\x00(\x00\x00\x00\x03\x04\x08\x00(\x00\x00\x00\x03\x02\x08\x00h\x00\x00\x00\x03\x01\x01\x0b\x00\x01\x00\x01\x02(\x02\x01\x03 \x14\x00\x00\x00\x00\x17\x04\x00\x00\x00\x00\x00\x00\x00\x16"

    out = to_smiles(bytestring)
    assert isinstance(out, str)
    assert out == "CC(=O)O"


def test_rdkit_mol_from_smiles_bad_smiles():
    """Test that a RuntimeError is raised on invalid smiles"""
    bad_smiles = ["not_a_smiles", "HHH", "C1.c2"]

    for smiles in bad_smiles:
        with pytest.raises(RuntimeError):
            rdkit_mol_from_smiles(smiles)


def test_rdkit_mol_from_smiles_bad_types():
    """Test that a TypeError is raised on non-string inputs"""
    bad_types_samples = [1, 3.14, [1, 2]]

    for sample in bad_types_samples:
        with pytest.raises(TypeError):
            rdkit_mol_from_smiles(sample)  # type: ignore


def test_hex_to_smiles():
    """That can convert a hex string to SMILES"""
    # Acetic Acid
    hexstring = "efbeadde000000000d0000000000000002000000040000000300000080010600600000000103060028000000030408002800000003020800680000000301010b00010001022802010320140000000017040000000000000016"

    out = to_smiles(hexstring)
    assert isinstance(out, str)
    assert out == "CC(=O)O"


def test_hex_to_oemol():
    """That can convert a hex string to OEMol object"""
    # Acetic Acid
    hexstring = "efbeadde000000000d0000000000000002000000040000000300000080010600600000000103060028000000030408002800000003020800680000000301010b00010001022802010320140000000017040000000000000016"

    out = to_rdkit_mol(hexstring)
    assert isinstance(out, Mol)

    out_as_smiles = to_smiles(out)
    assert out_as_smiles == "CC(=O)O"



import json
import os
from pathlib import Path

import pandas as pd
import pytest

from datasets import Dataset
from molflux.modelzoo import load_from_store, save_to_store
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.load import load_model
from molflux.modelzoo.models.sklearn.sklearn_pipeline.sklearn_pipeline_regressor import (
    SklearnPipelineRegressor,
)
from molflux.modelzoo.protocols import Estimator, Model
from molflux.modelzoo.typing import DataFrameLike

model_name = "sklearn_pipeline_regressor"

_X_FEATURES = ["X_col_1", "X_col_2", "X_col_3"]
_Y_FEATURES = ["y_col"]


@pytest.fixture(scope="function")
def fixture_basic_model() -> Model:
    return load_model(model_name, x_features=_X_FEATURES, y_features=_Y_FEATURES)


@pytest.fixture(scope="function")
def fixture_config_dir(fixture_path_to_assets: Path) -> Path:
    path: Path = (
        fixture_path_to_assets / "plugins" / "sklearn" / "sklearn_pipeline_regressor"
    )
    return path


def load_model_config(path_to_dir: Path, file_name: str) -> dict:
    path_to_file = os.path.join(path_to_dir, file_name)
    with open(path_to_file) as f:
        config: dict = json.load(f)
    return config


@pytest.fixture(scope="function")
def fixture_ridge_model(fixture_config_dir: Path) -> Model:
    file_name = "normalizer-kernel-ridge.json"
    model_config = load_model_config(fixture_config_dir, file_name)
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        **{"step_configs": model_config},
    )


@pytest.fixture(scope="function")
def fixture_svr_model(fixture_config_dir: Path) -> Model:
    file_name = "pca-select-k-svr.json"
    model_config = load_model_config(fixture_config_dir, file_name)
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        **{"step_configs": model_config},
    )


@pytest.fixture(scope="function")
def fixture_rf_model(fixture_config_dir: Path) -> Model:
    file_name = "scaler-random-forest.json"
    model_config = load_model_config(fixture_config_dir, file_name)
    return load_model(
        model_name,
        x_features=_X_FEATURES,
        y_features=_Y_FEATURES,
        **{"step_configs": model_config},
    )


@pytest.fixture(scope="function")
def fixture_train_dataset() -> DataFrameLike:
    train_df = pd.DataFrame(
        [
            [1, 1, 2, 10],
            [1, 1, 3, 11],
            [2, 3, 1, 13],
        ],
        columns=_X_FEATURES + _Y_FEATURES,
    )
    return Dataset.from_pandas(train_df)


@pytest.fixture(scope="function")
def fixture_predict_dataset() -> DataFrameLike:
    predict_df = pd.DataFrame(
        [
            [1, 2, 2],
            [1, 3, 1],
            [2, 5, 3],
        ],
        columns=_X_FEATURES,
    )
    return Dataset.from_pandas(predict_df)


@pytest.fixture(scope="function")
def fixture_empty_predict_dataset() -> DataFrameLike:
    predict_df = pd.DataFrame(
        [],
        columns=_X_FEATURES,
    )
    return Dataset.from_pandas(predict_df)


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_default_model_tag_matches_entrypoint_name(fixture_basic_model):
    """That the default model tag includes the catalogue entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_basic_model
    assert model_name in model.tag


@pytest.mark.parametrize(
    ["model_fixture", "expected_tag"],
    [
        [
            "fixture_basic_model",
            "sklearn_pipeline_regressor[sklearn.ensemble.RandomForestRegressor]",
        ],
        [
            "fixture_ridge_model",
            "sklearn_pipeline_regressor[sklearn.kernel_ridge.KernelRidge]",
        ],
        [
            "fixture_rf_model",
            "sklearn_pipeline_regressor[sklearn.ensemble.RandomForestRegressor]",
        ],
        ["fixture_svr_model", "sklearn_pipeline_regressor[sklearn.svm.SVR]"],
    ],
)
def test_tag_for_wrapped_model(model_fixture, expected_tag, request):
    """That the model has appropriate tag for a wrapped model.
    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = request.getfixturevalue(model_fixture)
    assert model.tag == expected_tag


@pytest.mark.parametrize(
    ["model_fixture"],
    [
        ["fixture_basic_model"],
        ["fixture_ridge_model"],
        ["fixture_rf_model"],
        ["fixture_svr_model"],
    ],
)
def test_is_mapped_to_correct_class(model_fixture, request):
    """That the model name is mapped to the appropriate class."""
    model = request.getfixturevalue(model_fixture)
    assert isinstance(model, SklearnPipelineRegressor)
    assert model_name in model.tag


@pytest.mark.parametrize(
    ["model_fixture"],
    [
        ["fixture_basic_model"],
        ["fixture_ridge_model"],
        ["fixture_rf_model"],
        ["fixture_svr_model"],
    ],
)
def test_implements_protocol(model_fixture, request):
    """That the model implements the public Estimator protocol."""
    model = request.getfixturevalue(model_fixture)
    assert isinstance(model, Estimator)


@pytest.mark.parametrize(
    ["model_fixture"],
    [
        ["fixture_basic_model"],
        ["fixture_ridge_model"],
        ["fixture_rf_model"],
        ["fixture_svr_model"],
    ],
)
def test_train_predict_model(
    model_fixture,
    fixture_train_dataset,
    fixture_predict_dataset,
    request,
):
    """Test that a model can run the train and predict functions"""

    model = request.getfixturevalue(model_fixture)
    train_dataset = fixture_train_dataset
    predict_dataset = fixture_predict_dataset

    model.train(train_dataset)
    predictions = model.predict(predict_dataset)
    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == len(predict_dataset)


@pytest.mark.parametrize(
    ["model_fixture"],
    [
        ["fixture_basic_model"],
        ["fixture_ridge_model"],
        ["fixture_rf_model"],
        ["fixture_svr_model"],
    ],
)
def test_predict_on_empty_dataset(
    model_fixture,
    fixture_train_dataset,
    fixture_empty_predict_dataset,
    request,
):
    """Test can predict on an empty input dataset.

    This should return an empty dictionary of output features.
    """

    model = request.getfixturevalue(model_fixture)
    train_dataset = fixture_train_dataset
    predict_dataset = fixture_empty_predict_dataset

    model.train(train_dataset)
    predictions = model.predict(predict_dataset)
    assert len(predictions) == len(model.y_features)
    for task_predictions in predictions.values():
        assert len(task_predictions) == 0  # == len(predict_dataset)


@pytest.mark.parametrize(
    ["model_fixture"],
    [
        ["fixture_basic_model"],
        ["fixture_ridge_model"],
        ["fixture_rf_model"],
        ["fixture_svr_model"],
    ],
)
def test_saving_loading(
    tmp_path,
    model_fixture,
    fixture_train_dataset,
    fixture_predict_dataset,
    request,
):
    """Test that a model can successfully be saved and loaded"""

    model = request.getfixturevalue(model_fixture)

    model.train(fixture_train_dataset)

    save_to_store(tmp_path, model)
    loaded_model = load_from_store(tmp_path)

    assert loaded_model.predict(fixture_predict_dataset)
